##
# Copyright 2009-2024 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for installing ANSYS optiSLang, implemented as an easyblock

@author: Chia-Jung Hsu (Chalmers University for Technology)
"""
import glob
import os
import stat
import tempfile

from easybuild.easyblocks.generic.packedbinary import PackedBinary
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.filetools import adjust_permissions
from easybuild.tools.run import run_cmd


class EB_optiSLang(PackedBinary):
    """Support for installing ANSYS optiSLang."""

    def __init__(self, *args, **kwargs):
        """Initialize optiSLang-specific variables."""
        super(EB_optiSLang, self).__init__(*args, **kwargs)

    def install_step(self):
        """Custom install procedure for ANSYS optiSLang."""
        licservs = self.cfg['license_server']
        if licservs is None:
            licservs = os.getenv('EB_ANSYS_LICENSE_SERVER', 'license.example.com')
        licservs = licservs.split(',')
        licport = self.cfg['license_server_port']
        if licport is None:
            licport = os.getenv('EB_ANSYS_LICENSE_SERVER_PORT', '2325:1055')
        licopts = ['-licserverinfo %s:%s' % (licport, serv) for serv in licservs]
        licoptsstr = ' '.join(licopts)

        tmpdir = tempfile.mkdtemp()

        # Sources (e.g. iso files) may drop the execute permissions
        adjust_permissions('INSTALL', stat.S_IXUSR)
        cmd = "./INSTALL -silent -install_dir %s -usetempdir %s %s" % (self.installdir, tmpdir, licoptsstr)
        run_cmd(cmd, log_all=True, simple=True)

        adjust_permissions(self.installdir, stat.S_IWOTH, add=False)

    def make_module_req_guess(self):
        """Custom extra module file entries for ANSYS optiSLang."""

        guesses = super(EB_optiSLang, self).make_module_req_guess()

        idirs = glob.glob(os.path.join(self.installdir, 'v*/'))
        if len(idirs) == 1:
            subdir = os.path.relpath(idirs[0], self.installdir)
            bindirs = [
                'aisol/bin/linx64',
                'dpf/bin/linux64',
                'optiSLang',
            ]
            guesses['PATH'] = [os.path.join(subdir, d) for d in bindirs]

        return guesses

    def sanity_check_step(self):
        """Custom sanity check for ANSYS optiSLang."""

        idirs = glob.glob(os.path.join(self.installdir, 'v*/'))
        if len(idirs) == 1:
            subdir = os.path.relpath(idirs[0], self.installdir)
        else:
            raise EasyBuildError("Failed to locate single install subdirectory v*/")

        errorlogfile = os.path.join(self.installdir, "install.err")
        if os.path.isfile(errorlogfile):
            raise EasyBuildError("Found error log from optiSLang installer %s" % errorlogfile)

        custom_paths = {
            'files': [os.path.join(subdir, 'optiSLang', x) for x in ['optislang', 'optislang-python']],
            'dirs': [os.path.join(subdir, x) for x in ['aisol', 'dpf']]
        }
        custom_commands = ['optislang --help']

        super(EB_optiSLang, self).sanity_check_step(custom_paths=custom_paths, custom_commands=custom_commands)



##
# Copyright 2018-2024 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for installing a bundle of Python packages, implemented as a generic easyblock

@author: Kenneth Hoste (Ghent University)
"""
import os
import sys

from easybuild.easyblocks.generic.bundle import Bundle
from easybuild.easyblocks.generic.pythonpackage import EBPYTHONPREFIXES, EXTS_FILTER_PYTHON_PACKAGES
from easybuild.easyblocks.generic.pythonpackage import PythonPackage, get_pylibdirs, pick_python_cmd
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.filetools import which
from easybuild.tools.modules import get_software_root
import easybuild.tools.environment as env


class PythonBundle(Bundle):
    """
    Bundle of PythonPackages: install Python packages as extensions in a bundle
    Defines custom sanity checks and module environment
    """

    @staticmethod
    def extra_options(extra_vars=None):
        """Easyconfig parameters specific to bundles of Python packages."""
        if extra_vars is None:
            extra_vars = {}
        # combine custom easyconfig parameters of Bundle & PythonPackage
        extra_vars = Bundle.extra_options(extra_vars)
        return PythonPackage.extra_options(extra_vars)

    def __init__(self, *args, **kwargs):
        """Initialize PythonBundle easyblock."""
        super(PythonBundle, self).__init__(*args, **kwargs)

        self.cfg['exts_defaultclass'] = 'PythonPackage'
        self.cfg['exts_filter'] = EXTS_FILTER_PYTHON_PACKAGES

        # need to disable templating to ensure that actual value for exts_default_options is updated...
        with self.cfg.disable_templating():
            # set default options for extensions according to relevant top-level easyconfig parameters
            pypkg_keys = PythonPackage.extra_options().keys()
            for key in pypkg_keys:
                if key not in self.cfg['exts_default_options']:
                    self.cfg['exts_default_options'][key] = self.cfg[key]

            self.cfg['exts_default_options']['download_dep_fail'] = True
            self.log.info("Detection of downloaded extension dependencies is enabled")

            self.log.info("exts_default_options: %s", self.cfg['exts_default_options'])

        self.pylibdir = None
        self.all_pylibdirs = []

        # figure out whether this bundle of Python packages is being installed for multiple Python versions
        self.multi_python = 'Python' in self.cfg['multi_deps']

    def prepare_step(self, *args, **kwargs):
        """Prepare for installing bundle of Python packages."""
        super(Bundle, self).prepare_step(*args, **kwargs)

        python_root = get_software_root('Python')
        if python_root is None:
            raise EasyBuildError("Python not included as dependency!")

        # when system Python is used, the first 'python' command in $PATH will not be $EBROOTPYTHON/bin/python,
        # since $EBROOTPYTHON is set to just 'Python' in that case
        # (see handling of allow_system_deps in EasyBlock.prepare_step)
        if which('python') == os.path.join(python_root, 'bin', 'python'):
            # if we're using a proper Python dependency, let det_pylibdir use 'python' like it does by default
            python_cmd = None
        else:
            # since det_pylibdir will use 'python' by default as command to determine Python lib directory,
            # we need to intervene when the system Python is used, by specifying version requirements
            # to pick_python_cmd so the right 'python' command is used;
            # if we're using the system Python and no Python version requirements are specified,
            # use major/minor version of Python being used in this EasyBuild session (as we also do in PythonPackage)
            req_py_majver = self.cfg['req_py_majver']
            if req_py_majver is None:
                req_py_majver = sys.version_info[0]
            req_py_minver = self.cfg['req_py_minver']
            if req_py_minver is None:
                req_py_minver = sys.version_info[1]

            python_cmd = pick_python_cmd(req_maj_ver=req_py_majver, req_min_ver=req_py_minver)

            # If pick_python_cmd didn't find a (system) Python command, we should raise an error
            if python_cmd:
                self.log.info("Python command being used: %s", python_cmd)
            else:
                raise EasyBuildError(
                    "Failed to pick Python command that satisfies requirements in the easyconfig "
                    "(req_py_majver = %s, req_py_minver = %s)", req_py_majver, req_py_minver
                )

        self.all_pylibdirs = get_pylibdirs(python_cmd=python_cmd)
        self.pylibdir = self.all_pylibdirs[0]

        # if 'python' is not used, we need to take that into account in the extensions filter
        # (which is also used during the sanity check)
        if python_cmd:
            orig_exts_filter = EXTS_FILTER_PYTHON_PACKAGES
            self.cfg['exts_filter'] = (orig_exts_filter[0].replace('python', python_cmd), orig_exts_filter[1])

    def extensions_step(self, *args, **kwargs):
        """Install extensions (usually PythonPackages)"""
        # don't add user site directory to sys.path (equivalent to python -s)
        env.setvar('PYTHONNOUSERSITE', '1', verbose=False)
        super(PythonBundle, self).extensions_step(*args, **kwargs)

    def test_step(self):
        """No global test step for bundle of Python packages."""
        # required since runtest is set to True for Python packages by default
        pass

    def make_module_extra(self, *args, **kwargs):
        """Extra statements to include in module file: update $PYTHONPATH."""
        txt = super(Bundle, self).make_module_extra(*args, **kwargs)

        # update $EBPYTHONPREFIXES rather than $PYTHONPATH
        # if this Python package was installed for multiple Python versions
        if self.multi_python:
            txt += self.module_generator.prepend_paths(EBPYTHONPREFIXES, '')
        else:

            # the temporary module file that is generated before installing extensions
            # must add all subdirectories to $PYTHONPATH without checking existence,
            # otherwise paths will be missing since nothing is there initially
            if self.current_step == 'extensions':
                new_pylibdirs = self.all_pylibdirs
            else:
                new_pylibdirs = [
                    lib_dir for lib_dir in self.all_pylibdirs
                    if os.path.exists(os.path.join(self.installdir, lib_dir))
                ]

            for pylibdir in new_pylibdirs:
                txt += self.module_generator.prepend_paths('PYTHONPATH', pylibdir)

        return txt

    def load_module(self, *args, **kwargs):
        """
        Make sure that $PYTHONNOUSERSITE is defined after loading module file for this software."""

        super(PythonBundle, self).load_module(*args, **kwargs)

        # Don't add user site directory to sys.path (equivalent to python -s),
        # to avoid that any Python packages installed in $HOME/.local/lib affect the sanity check.
        # Required here to ensure that it is defined for sanity check commands of the bundle
        # because the environment is reset to the initial environment right before loading the module
        env.setvar('PYTHONNOUSERSITE', '1', verbose=False)

    def sanity_check_step(self, *args, **kwargs):
        """Custom sanity check for bundle of Python package."""

        # inject directory path that uses %(pyshortver)s template into default value for sanity_check_paths
        # this is relevant for installations of Python bundles for multiple Python versions (via multi_deps)
        # (we can not pass this via custom_paths, since then the %(pyshortver)s template value will not be resolved)
        if not self.cfg['sanity_check_paths']:
            self.cfg['sanity_check_paths'] = {
                'files': [],
                'dirs': [os.path.join('lib', 'python%(pyshortver)s', 'site-packages')],
            }

        super(Bundle, self).sanity_check_step(*args, **kwargs)



##
# Copyright 2013-2024 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for building and installing ESMF, implemented as an easyblock

@author: Kenneth Hoste (Ghent University)
@author: Damian Alvarez (Forschungszentrum Juelich GmbH)
@author: Maxime Boissonneault (Digital Research Alliance of Canada)
"""
import os
from easybuild.tools import LooseVersion

import easybuild.tools.environment as env
import easybuild.tools.toolchain as toolchain
from easybuild.easyblocks.generic.configuremake import ConfigureMake
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.modules import get_software_root, get_software_version
from easybuild.tools.run import run_cmd
from easybuild.tools.systemtools import get_shared_lib_ext
from easybuild.framework.easyconfig import CUSTOM


class EB_ESMF(ConfigureMake):
    """Support for building/installing ESMF."""

    @staticmethod
    def extra_options():
        """Custom easyconfig parameters for ESMF"""
        extra_vars = {
            'disable_lapack': [False, 'Disable external LAPACK - True or False', CUSTOM]
        }
        return ConfigureMake.extra_options(extra_vars)

    def configure_step(self):
        """Custom configuration procedure for ESMF through environment variables."""

        env.setvar('ESMF_DIR', self.cfg['start_dir'])
        env.setvar('ESMF_INSTALL_PREFIX', self.installdir)
        env.setvar('ESMF_INSTALL_BINDIR', 'bin')
        env.setvar('ESMF_INSTALL_LIBDIR', 'lib')
        env.setvar('ESMF_INSTALL_MODDIR', 'mod')

        # specify compiler
        comp_family = self.toolchain.comp_family()
        if comp_family in [toolchain.GCC]:
            compiler = 'gfortran'
        else:
            compiler = comp_family.lower()
        env.setvar('ESMF_COMPILER', compiler)

        env.setvar('ESMF_F90COMPILEOPTS', os.getenv('F90FLAGS'))
        env.setvar('ESMF_CXXCOMPILEOPTS', os.getenv('CXXFLAGS'))

        # specify MPI communications library
        comm = None
        mpi_family = self.toolchain.mpi_family()
        if mpi_family in [toolchain.MPICH, toolchain.QLOGICMPI]:
            # MPICH family for MPICH v3.x, which is MPICH2 compatible
            comm = 'mpich2'
        else:
            comm = mpi_family.lower()
        env.setvar('ESMF_COMM', comm)

        # specify decent LAPACK lib
        if self.cfg['disable_lapack']:
            env.setvar('ESMF_LAPACK', 'OFF')
            # MOAB can't be built without LAPACK
            env.setvar('ESMF_MOAB', 'OFF')
        else:
            env.setvar('ESMF_LAPACK', 'user')
            ldflags = os.getenv('LDFLAGS')
            liblapack = os.getenv('LIBLAPACK_MT') or os.getenv('LIBLAPACK')
            if liblapack is None:
                msg = "$LIBLAPACK(_MT) not defined, no BLAS/LAPACK in %s toolchain?"
                raise EasyBuildError(msg, self.toolchain.name)
            else:
                env.setvar('ESMF_LAPACK_LIBS', ldflags + ' ' + liblapack)

        # specify netCDF
        netcdf = get_software_root('netCDF')
        if netcdf:
            if LooseVersion(self.version) >= LooseVersion('7.1.0'):
                env.setvar('ESMF_NETCDF', 'nc-config')
            else:
                env.setvar('ESMF_NETCDF', 'user')
                netcdf_libs = ['-L%s/lib' % netcdf, '-lnetcdf']

                # Fortran
                netcdff = get_software_root('netCDF-Fortran')
                if netcdff:
                    netcdf_libs = ["-L%s/lib" % netcdff] + netcdf_libs + ["-lnetcdff"]
                else:
                    netcdf_libs.append('-lnetcdff')

                # C++
                netcdfcxx = get_software_root('netCDF-C++')
                if netcdfcxx:
                    netcdf_libs = ["-L%s/lib" % netcdfcxx] + netcdf_libs + ["-lnetcdf_c++"]
                else:
                    netcdfcxx = get_software_root('netCDF-C++4')
                    if netcdfcxx:
                        netcdf_libs = ["-L%s/lib" % netcdfcxx] + netcdf_libs + ["-lnetcdf_c++4"]
                    else:
                        netcdf_libs.append('-lnetcdf_c++')
                env.setvar('ESMF_NETCDF_LIBS', ' '.join(netcdf_libs))

        # 'make info' provides useful debug info
        cmd = "make info"
        run_cmd(cmd, log_all=True, simple=True, log_ok=True)

    def install_step(self):
        # first, install the software
        super(EB_ESMF, self).install_step()

        python = get_software_version('Python')
        if python:
            # then, install the python bindings
            py_subdir = os.path.join(self.builddir, 'esmf-ESMF_%s' % '_'.join(self.version.split('.')),
                                     'src', 'addon', 'ESMPy')
            try:
                os.chdir(py_subdir)
            except OSError as err:
                raise EasyBuildError("Failed to move to: %s", err)

            cmd = "python setup.py build --ESMFMKFILE=%s/lib/esmf.mk " % self.installdir
            cmd += " && python setup.py install --prefix=%s" % self.installdir
            run_cmd(cmd, log_all=True, simple=True, log_ok=True)

    def make_module_extra(self):
        """Add install path to PYTHONPATH or EBPYTHONPREFIXES"""
        txt = super(EB_ESMF, self).make_module_extra()

        # set environment variable ESMFMKFILE
        # see section 9.9 in https://earthsystemmodeling.org/docs/release/latest/ESMF_usrdoc/node10.html
        esmf_mkfile_path = os.path.join(self.installdir, "lib", "esmf.mk")
        txt += self.module_generator.set_environment('ESMFMKFILE', esmf_mkfile_path)

        if self.cfg['multi_deps'] and 'Python' in self.cfg['multi_deps']:
            txt += self.module_generator.prepend_paths('EBPYTHONPREFIXES', '')
        else:
            python = get_software_version('Python')
            if python:
                pyshortver = '.'.join(get_software_version('Python').split('.')[:2])
                pythonpath = os.path.join('lib', 'python%s' % pyshortver, 'site-packages')
                txt += self.module_generator.prepend_paths('PYTHONPATH', [pythonpath])

        return txt

    def sanity_check_step(self):
        """Custom sanity check for ESMF."""

        if LooseVersion(self.version) < LooseVersion('8.1.0'):
            binaries = ['ESMF_Info', 'ESMF_InfoC', 'ESMF_Regrid', 'ESMF_RegridWeightGen',
                        'ESMF_Scrip2Unstruct', 'ESMF_WebServController']
        else:
            binaries = ['ESMF_PrintInfo', 'ESMF_PrintInfoC', 'ESMF_Regrid', 'ESMF_RegridWeightGen',
                        'ESMF_Scrip2Unstruct', 'ESMF_WebServController']

        libs = ['esmf.mk', 'libesmf.a', 'libesmf.%s' % get_shared_lib_ext()]
        custom_paths = {
            'files': [os.path.join('bin', x) for x in binaries] + [os.path.join('lib', x) for x in libs],
            'dirs': ['include', 'mod'],
        }

        custom_commands = []
        if get_software_root('Python'):
            custom_commands += ["python -c 'import ESMF'"]

        super(EB_ESMF, self).sanity_check_step(custom_commands=custom_commands, custom_paths=custom_paths)



##
# Copyright 2015-2024 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for building and installing OCaml + opam (+ extensions), implemented as an easyblock

@author: Kenneth Hoste (Ghent University)
"""
import os
import re
from easybuild.tools import LooseVersion

from easybuild.easyblocks.generic.configuremake import ConfigureMake
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.filetools import change_dir
from easybuild.tools.run import run_cmd


EXTS_FILTER_OCAML_PACKAGES = ("eval `opam config env` && opam list --installed %(ext_name)s.%(ext_version)s", '')
OPAM_SUBDIR = 'opam'


def det_opam_version():
    """Determine OPAM version (using 'opam --version')."""

    opam_ver = None

    opam_version_cmd = 'opam --version'
    (out, ec) = run_cmd(opam_version_cmd)
    if ec == 0:
        res = re.search('^[0-9.]+$', out.strip())
        if res:
            opam_ver = res.group(0)

    if opam_ver is None:
        raise EasyBuildError("Failed to determine OPAM version using '%s'!", opam_version_cmd)

    return opam_ver


def mk_opam_init_cmd(root=None):
    """Construct 'opam init' command."""

    opam_init_cmd = ['opam', 'init']

    if LooseVersion(det_opam_version()) >= LooseVersion('2.0.0'):
        # disable sandboxing, required bubblewrap (which requires setuid)
        # see http://opam.ocaml.org/doc/FAQ.html#Why-does-opam-require-bwrap
        opam_init_cmd.append('--disable-sandboxing')

    if root:
        opam_init_cmd.extend(['--root', root])

    return ' '.join(opam_init_cmd)


class EB_OCaml(ConfigureMake):
    """Support for building/installing OCaml + opam (+ additional extensions)."""

    def __init__(self, *args, **kwargs):
        """Initialisation of custom class variables for OCaml."""
        super(EB_OCaml, self).__init__(*args, **kwargs)
        self.with_opam = False

    def configure_step(self):
        """Custom configuration procedure for OCaml."""
        self.cfg['prefix_opt'] = '-prefix '
        if LooseVersion(self.version) < LooseVersion("4.12"):
            self.cfg.update('configopts', '-cc "%s %s"' % (os.environ['CC'], os.environ['CFLAGS']))

        if 'world.opt' not in self.cfg['buildopts']:
            self.cfg.update('buildopts', 'world.opt')

        super(EB_OCaml, self).configure_step()

    def install_step(self):
        """
        Custom install procedure for OCaml.
        First install OCaml using 'make install', then install OPAM (if sources are provided).
        """
        super(EB_OCaml, self).install_step()

        try:
            all_dirs = os.listdir(self.builddir)
        except OSError as err:
            raise EasyBuildError("Failed to check contents of %s: %s", self.builddir, err)

        opam_dirs = [d for d in all_dirs if d.startswith('opam')]
        if len(opam_dirs) == 1:
            # load temporary module so OCaml installation is available for building & installing opam
            fake_mod_data = self.load_fake_module()

            opam_dir = os.path.join(self.builddir, opam_dirs[0])
            self.log.info("Found unpacked OPAM sources at %s, so installing it.", opam_dir)
            self.with_opam = True
            change_dir(opam_dir)

            run_cmd("./configure --prefix=%s" % self.installdir)
            run_cmd("make lib-ext")  # locally build/install required dependencies
            run_cmd("make")
            run_cmd("make install")

            opam_init_cmd = mk_opam_init_cmd(root=os.path.join(self.installdir, OPAM_SUBDIR))
            run_cmd(opam_init_cmd)

            self.clean_up_fake_module(fake_mod_data)
        else:
            self.log.warning("OPAM sources not found in %s: %s", self.builddir, all_dirs)

    def prepare_for_extensions(self):
        """Set default class and filter for OCaml packages."""
        # build and install additional packages with OCamlPackage easyblock
        self.cfg['exts_defaultclass'] = "OCamlPackage"
        self.cfg['exts_filter'] = EXTS_FILTER_OCAML_PACKAGES
        super(EB_OCaml, self).prepare_for_extensions()

    def collect_exts_file_info(self, *args, **kwargs):
        """Don't fetch extension sources, OPAM takes care of that (and archiving too)."""
        return [{'name': ext_name, 'version': ext_version} for ext_name, ext_version in self.cfg['exts_list']]

    def sanity_check_step(self):
        """Custom sanity check for OCaml."""
        binaries = ['bin/ocaml', 'bin/ocamlc', 'bin/ocamlopt', 'bin/ocamlrun']
        dirs = []
        if self.with_opam:
            binaries.append('bin/opam')
            dirs.append(OPAM_SUBDIR)

        extension_names = [ext_name for ext_name, _ in self.cfg['exts_list']]

        custom_commands = ["ocaml --help"]
        if 'ocamlfind' in extension_names:
            custom_commands.append("ocamlfind list")

        if 'dune' in extension_names:
            custom_commands.append("dune --version")

        custom_paths = {
            'files': binaries,
            'dirs': dirs,
        }

        super(EB_OCaml, self).sanity_check_step(custom_paths=custom_paths, custom_commands=custom_commands)

    def make_module_req_guess(self):
        """Custom extra paths/variables to define in generated module for OCaml."""
        guesses = super(EB_OCaml, self).make_module_req_guess()

        guesses.update({
            'CAML_LD_LIBRARY_PATH': ['lib'],
            'OPAMROOT': [OPAM_SUBDIR],
            'PATH': ['bin', os.path.join(OPAM_SUBDIR, 'default', 'bin')],
        })

        return guesses



##
# Copyright 2009-2024 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for building and installing g2clib, implemented as an easyblock

@author: Stijn De Weirdt (Ghent University)
@author: Dries Verdegem (Ghent University)
@author: Kenneth Hoste (Ghent University)
@author: Pieter De Baets (Ghent University)
@author: Jens Timmerman (Ghent University)
"""

import glob
import os
import shutil

from easybuild.easyblocks.generic.configuremake import ConfigureMake
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.modules import get_software_root

from easybuild.tools import LooseVersion


class EB_g2clib(ConfigureMake):
    """Support for building g2clib GRIB2 C library."""

    def configure_step(self):
        """No configuration needed"""
        pass

    def build_step(self):
        """Build by supplying required make options, and running build_step."""

        jasper = get_software_root('JASPER')
        if not jasper:
            raise EasyBuildError("JasPer module not loaded?")

        # beware: g2clib uses INC, while g2lib uses INCDIR !
        buildopts = ' '.join([
            r'CC="%s"' % os.getenv('CC'),
            r'CFLAGS="%s \$(INC) \$(DEF) -D__64BIT__"' % os.getenv('CFLAGS'),
            r'FC="%s"' % os.getenv('F90'),
            r'INC="-I%s"' % os.path.join(jasper, 'include'),
        ])
        self.cfg.update('buildopts', buildopts)

        super(EB_g2clib, self).build_step()

    def install_step(self):
        """Install by copying library and header files to install directory."""

        try:
            # copy library
            targetdir = os.path.join(self.installdir, "lib")
            os.mkdir(targetdir)

            # not documented AFAIK, but real
            if LooseVersion(self.version) < LooseVersion('1.6.0'):
                fn = "libgrib2c.a"
            else:
                fn = "libg2c_v%s.a" % self.version

            shutil.copyfile(os.path.join(self.cfg['start_dir'], fn),
                            os.path.join(targetdir, fn))

            # make link so other software still finds libgrib2c.a
            if LooseVersion(self.version) >= LooseVersion('1.6.0'):
                os.symlink("libg2c_v%s.a" % self.version, os.path.join(targetdir, "libgrib2c.a"))

            # copy header files
            targetdir = os.path.join(self.installdir, "include")
            os.mkdir(targetdir)
            for fn in glob.glob('*.h'):
                shutil.copyfile(os.path.join(self.cfg['start_dir'], fn),
                                os.path.join(targetdir, fn))

        except OSError as err:
            raise EasyBuildError("Failed to copy files to install dir: %s", err)

    def sanity_check_step(self):
        """Custom sanity check for g2clib."""

        custom_paths = {
            'files': ["lib/libgrib2c.a"],
            'dirs': ["include"]
        }

        super(EB_g2clib, self).sanity_check_step(custom_paths=custom_paths)



##
# Copyright 2009-2023 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for `metalwalls`, implemented as an easyblock

@author: Davide Grassano (CECAM, EPFL)
"""
import os

import easybuild.tools.environment as env
import easybuild.tools.toolchain as toolchain
from easybuild.tools import LooseVersion
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.modules import get_software_root, get_software_version
from easybuild.tools.run import run_cmd

from easybuild.easyblocks.generic.makecp import MakeCp


class EB_MetalWalls(MakeCp):
    """Support for building and installing `metalwalls`."""

    def __init__(self, *args, **kwargs):
        """Add extra config options specific to `metalwalls`."""
        super(EB_MetalWalls, self).__init__(*args, **kwargs)

        self._build_python_interface = False
        self._extra_schecks = []

    def configure_step(self):
        """Custom configuration procedure for `metalwalls`."""
        comp_fam = self.toolchain.comp_family()

        f90flags = os.getenv('F90FLAGS', '').split(' ')
        fppflags = os.getenv('FPPFLAGS', '').split(' ')
        ldflags = os.getenv('LDFLAGS', '').split(' ')
        f90wrap = ''
        f2py = ''
        fcompiler = ''

        if comp_fam == toolchain.INTELCOMP:
            jflag = '-module'
            fppflags = ['-fpp']
        elif comp_fam == toolchain.GCC:
            jflag = '-J'
            fppflags = ['-cpp']
        elif comp_fam == toolchain.NVHPC:
            if self.toolchain.options.get('usempi', False):
                raise EasyBuildError('NVHPC compilation does not support MPI')
            jflag = '-module '  # Space is needed
            fppflags = ['-DMW_SERIAL', '-Mpreprocess']
            f90flags += ['-Minline', '-acc', '-gpu=managed']
        else:
            raise EasyBuildError('Unsupported compiler family: %s' % comp_fam)

        # https://gitlab.com/ampere2/metalwalls/-/wikis/install#plumed
        plumed = get_software_root('PLUMED')
        f90wrap = get_software_root('f90wrap')
        f90wrap_version = get_software_version('f90wrap')

        if LooseVersion(self.version) <= LooseVersion('21.06.1'):
            if LooseVersion(f90wrap_version) > LooseVersion('0.2.13'):
                raise EasyBuildError('MetalWalls version %s requires f90wrap <= 0.2.13' % self.version)

        tpl_rgx = 'alltests\\.append(suite_%s)'
        if plumed:
            f90flags += ['-fallow-argument-mismatch']  # Code inside ifdef causes mismatch errors
            fppflags += ['-DMW_USE_PLUMED']
            cmd = ['touch', 'mw2.diff']
            run_cmd(' '.join(cmd), log_all=False, log_ok=False, simple=False, regexp=False)
            cmd = ['plumed', 'patch', '-d mw2.diff', '--patch', '--shared', '--engine', 'mw2']
            run_cmd(' '.join(cmd), log_all=True, simple=False)
        else:
            self.log.info('PLUMED not found, excluding from test-suite')
            rgx = tpl_rgx % 'plumed'
            cmd = ['sed', '-i', "'s/^\\( \\+\\)%s$/\\1pass # %s/'" % (rgx, rgx), 'tests/regression_tests.py']
            cmd = ['sed', '-i', "'s/%s/pass/'" % rgx, 'tests/regression_tests.py']
            run_cmd(' '.join(cmd), log_all=True, simple=False)

        if f90wrap:
            if not get_software_root('mpi4py'):
                raise EasyBuildError('Building the Python interface requires mpi4py')
            self._build_python_interface = True
            f90wrap = 'f90wrap'
            f2py = 'f2py'
            fcompiler = os.getenv('F90_SEQ')
            f90flags += ['-fPIC']
            self.cfg.update('build_cmd_targets', 'python')
            self.cfg.update('files_to_copy', 'build/python')
            self._extra_schecks += ['python/mw.py', 'python/metalwalls.py']

        else:
            self.log.info('f90wrap not found, excluding python interface from test-suite')
            rgx = tpl_rgx % 'python_interface'
            cmd = ['sed', '-i', "'s/%s/pass/'" % rgx, 'tests/regression_tests.py']
            run_cmd(' '.join(cmd), log_all=True, simple=False)

        # Add libraries with LAPACK support
        lapack_shared_libs = os.getenv('LAPACK_SHARED_LIBS', None)
        if not lapack_shared_libs:
            raise EasyBuildError('Must use a toolchain with LAPACK support')
        extra_libs = lapack_shared_libs.replace('.so', '').split(',')
        extra_libs = [_[3:] for _ in extra_libs if _.startswith('lib')]
        ldflags += ['-l%s' % lib for lib in extra_libs]

        # Write the `config.mk` file
        configmk = os.path.join(self.cfg['start_dir'], 'config.mk')
        with open(configmk, 'w') as f:
            f.write('F90FLAGS := %s\n' % ' '.join(f90flags))
            f.write('FPPFLAGS := %s\n' % ' '.join(fppflags))
            f.write('LDFLAGS := %s\n' % ' '.join(ldflags))
            f.write('F90WRAP := %s\n' % f90wrap)
            f.write('F2PY := %s\n' % f2py)
            f.write('FCOMPILER := %s\n' % fcompiler)
            f.write('J := %s\n' % jflag)

        with open(configmk, 'r') as f:
            self.log.info('Contents of generated `config.mk`:\n%s' % f.read())

    def test_step(self):
        """
        Test the compilation using `metalwalls`'s test suite.
        """

        if self._build_python_interface:
            ppath = os.getenv('PYTHONPATH', '')
            ppath = os.path.join(self.cfg['start_dir'], 'build', 'python') + ':' + ppath
            self.log.info('Setting PYTHONPATH for testing to %s' % ppath)
            env.setvar('PYTHONPATH', ppath)

        super(EB_MetalWalls, self).test_step()

    def make_module_extra(self, extra=None):
        """Add custom entries to module."""

        txt = super(EB_MetalWalls, self).make_module_extra()

        if self._build_python_interface:
            txt += self.module_generator.prepend_paths('PYTHONPATH', 'python')

        return txt

    def sanity_check_step(self):
        """Custom sanity check for Quantum ESPRESSO."""

        bins = ['mw']

        custom_paths = {
            'files': [os.path.join('bin', x) for x in bins] + self._extra_schecks,
            'dirs': []
        }

        super(EB_MetalWalls, self).sanity_check_step(custom_paths=custom_paths)



##
# Copyright 2009-2024 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for building and installing GAMESS-US, implemented as an easyblock

@author: Stijn De Weirdt (Ghent University)
@author: Dries Verdegem (Ghent University)
@author: Kenneth Hoste (Ghent University)
@author: Pieter De Baets (Ghent University)
@author: Jens Timmerman (Ghent University)
@author: Toon Willems (Ghent University)
@author: Pablo Escobar (sciCORE, SIB, University of Basel)
@author: Benjamin Roberts (The University of Auckland)
@author: Alex Domingo (Vrije Universiteit Brussel)
"""
import fileinput
import glob
import os
import re
import sys
import tempfile

from easybuild.framework.easyblock import EasyBlock
from easybuild.framework.easyconfig import CUSTOM, MANDATORY
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.config import build_option
from easybuild.tools.filetools import change_dir, copy_file, mkdir, read_file, write_file, remove_dir
from easybuild.tools.modules import get_software_root, get_software_version
from easybuild.tools.run import run_cmd
from easybuild.tools.systemtools import POWER, X86_64
from easybuild.tools.systemtools import get_cpu_architecture
from easybuild.tools import LooseVersion, toolchain

GAMESS_INSTALL_INFO = 'install.info'
GAMESS_SERIAL_TESTS = [
    'exam05',  # only the gradients for CITYP=CIS run in parallel
    'exam32',  # only CCTYP=CCSD or CCTYP=CCSD(T) can run in parallel
    'exam42',  # ROHF'S CCTYP must be CCSD or CR-CCL, with serial execution
    'exam45',  # only CCTYP=CCSD or CCTYP=CCSD(T) can run in parallel
    'exam46',  # ROHF'S CCTYP must be CCSD or CR-CCL, with serial execution
    'exam47',  # ROHF'S CCTYP must be CCSD or CR-CCL, with serial execution
]


class EB_GAMESS_minus_US(EasyBlock):
    """Support for building/installing GAMESS-US."""

    @staticmethod
    def extra_options():
        """Define custom easyconfig parameters for GAMESS-US."""
        extra_vars = {
            'ddi_comm': ['mpi', "DDI communication layer to use", CUSTOM],
            'maxcpus': [None, "Maximum number of cores per node", MANDATORY],
            'maxnodes': [None, "Maximum number of nodes", MANDATORY],
            'hyperthreading': [True, "Enable support for hyperthreading (2 threads per core)", CUSTOM],
            'runtest': [True, "Run GAMESS-US tests", CUSTOM],
            'scratch_dir': ['$TMPDIR', "dir for temporary binary files", CUSTOM],
            'user_scratch_dir': ['$TMPDIR', "dir for supplementary output files", CUSTOM],
        }
        return EasyBlock.extra_options(extra_vars)

    def __init__(self, *args, **kwargs):
        """Easyblock constructor, enable building in installation directory."""
        super(EB_GAMESS_minus_US, self).__init__(*args, **kwargs)
        self.build_in_installdir = True

        # resolve path to scratch dir and make it
        scratch_dir_resolved = os.path.expandvars(self.cfg['scratch_dir'])
        if "$" in scratch_dir_resolved:
            error_msg = "Provided scratch directory '%s' does not resolve into a path."
            raise EasyBuildError(error_msg, self.cfg['scratch_dir'])

        mkdir(scratch_dir_resolved, parents=True)

        self.testdir = None
        if self.cfg['runtest']:
            # run tests in scratch dir
            self.testdir = tempfile.mkdtemp(dir=scratch_dir_resolved)

            # make sure test dir doesn't contain [ or ], rungms csh script doesn't handle that well ("set: No match")
            if re.search(r'[\[\]]', self.testdir):
                error_msg = "Temporary dir for tests '%s' will cause problems with rungms csh script"
                raise EasyBuildError(error_msg, self.testdir)

        # OpenMP cannot be enabled across the board, activate it only through
        # GMS_OPENMP so that it gets enabled by the build where needed
        self.omp_enabled = False
        if self.toolchain.options.get('openmp', None):
            self.omp_enabled = True
            self.toolchain.options['openmp'] = False

    def extract_step(self):
        """Extract sources."""
        # strip off 'gamess' part to avoid having everything in a 'gamess' subdirectory
        self.cfg['unpack_options'] = "--strip-components=1"
        super(EB_GAMESS_minus_US, self).extract_step()

    def configure_step(self):
        """Configure GAMESS-US via install.info file"""
        installinfo_opts = {}

        # installation paths
        installinfo_opts["GMS_PATH"] = self.installdir
        installinfo_opts["GMS_BUILD_DIR"] = self.builddir

        # machine type
        cpu_arch = get_cpu_architecture()
        if cpu_arch == X86_64:
            machinetype = "linux64"
        elif cpu_arch == POWER:
            machinetype = "ibm64"
        else:
            raise EasyBuildError("Build target %s currently unsupported", cpu_arch)

        installinfo_opts["GMS_TARGET"] = machinetype
        installinfo_opts["GMS_HPC_SYSTEM_TARGET"] = "generic"

        # Compiler config
        comp_fam = self.toolchain.comp_family()
        fortran_comp, fortran_version = None, None
        if comp_fam == toolchain.INTELCOMP:
            fortran_comp = 'ifort'
            (out, _) = run_cmd("ifort -v", simple=False)
            res = re.search(r"^ifort version ([0-9]+)\.[0-9.]+$", out)
            try:
                version_num = res.group(1)
            except (AttributeError, IndexError):
                raise EasyBuildError("Failed to determine ifort major version number")
            fortran_version = {"GMS_IFORT_VERNO": version_num}
        elif comp_fam == toolchain.GCC:
            fortran_comp = 'gfortran'
            version_num = '.'.join(get_software_version('GCC').split('.')[:2])
            fortran_version = {"GMS_GFORTRAN_VERNO": version_num}
        else:
            raise EasyBuildError("Compiler family '%s' currently unsupported.", comp_fam)

        installinfo_opts["GMS_FORTRAN"] = fortran_comp
        installinfo_opts.update(fortran_version)

        # OpenMP config
        installinfo_opts["GMS_OPENMP"] = self.omp_enabled

        # Math library config
        known_mathlibs = ['imkl', 'OpenBLAS', 'ATLAS', 'ACML']
        loaded_mathlib = [mathlib for mathlib in known_mathlibs if get_software_root(mathlib)]

        # math library: default settings
        try:
            mathlib = loaded_mathlib[0].lower()
            mathlib_root = get_software_root(loaded_mathlib[0])
            mathlib_subfolder = ''
            mathlib_flags = ''
        except IndexError:
            raise EasyBuildError("None of the known math libraries (%s) available, giving up.", known_mathlibs)

        # math library: special cases
        if mathlib == 'imkl':
            mathlib = 'mkl'
            mathlib_subfolder = 'mkl/lib/intel64'
            mkl_version = '12'
            imkl_ver = get_software_version('imkl')
            if LooseVersion(imkl_ver) >= LooseVersion("2021"):
                # OneAPI version
                mathlib_subfolder = 'mkl/latest/lib/intel64'
                mkl_version = 'oneapi'
            installinfo_opts["GMS_MKL_VERNO"] = mkl_version

            if LooseVersion(self.version) >= LooseVersion('20230601'):
                installinfo_opts["GMS_THREADED_BLAS"] = self.omp_enabled

        elif mathlib == 'openblas':
            mathlib_flags = "-lopenblas -lgfortran"
            if LooseVersion(self.version) >= LooseVersion('20210101'):
                mathlib_subfolder = 'lib'

        if mathlib_root is not None:
            mathlib_path = os.path.join(mathlib_root, mathlib_subfolder)
            self.log.debug("Software root of math libraries set to: %s", mathlib_path)
        else:
            raise EasyBuildError("Software root of math libraries (%s) not found", mathlib)

        installinfo_opts["GMS_MATHLIB"] = mathlib
        installinfo_opts["GMS_MATHLIB_PATH"] = mathlib_path
        installinfo_opts["GMS_LAPACK_LINK_LINE"] = '"%s"' % mathlib_flags

        # verify selected DDI communication layer
        known_ddi_comms = ['mpi', 'mixed', 'shmem', 'sockets']
        if not self.cfg['ddi_comm'] in known_ddi_comms:
            raise EasyBuildError(
                "Unsupported DDI communication layer specified (known: %s): %s", known_ddi_comms, self.cfg['ddi_comm']
            )

        installinfo_opts["GMS_DDI_COMM"] = self.cfg['ddi_comm']

        # MPI library config
        mpilib, mpilib_path = '', ''
        if self.cfg['ddi_comm'] == 'mpi':
            known_mpilibs = ['impi', 'OpenMPI', 'MVAPICH2', 'MPICH2']
            loaded_mpilib = [mpilib for mpilib in known_mpilibs if get_software_root(mpilib)]

            # mpi library: default settings
            try:
                mpilib = loaded_mpilib[0].lower()
                mpilib_root = get_software_root(loaded_mpilib[0])
                mpilib_subfolder = ''
            except IndexError:
                raise EasyBuildError("None of the known MPI libraries (%s) available, giving up.", known_mpilibs)

            if mpilib == 'impi':
                impi_ver = get_software_version('impi')
                if LooseVersion(impi_ver) >= LooseVersion("2021"):
                    mpilib_subfolder = "mpi/latest"
                else:
                    mpilib_subfolder = "intel64"

                if LooseVersion(impi_ver) >= LooseVersion("2019"):
                    # fix rungms settings for newer versions of Intel MPI
                    print("PATCH IMPI")
                    rungms = os.path.join(self.builddir, 'rungms')
                    try:
                        for line in fileinput.input(rungms, inplace=1, backup='.orig'):
                            line = re.sub(r"^(\s*setenv\s*I_MPI_STATS).*", r"# \1", line)
                            line = re.sub(r"^(\s*setenv\s*I_MPI_WAIT_MODE)\s*enable.*", r"\1 1", line)
                            sys.stdout.write(line)
                    except IOError as err:
                        raise EasyBuildError("Failed to patch Intel MPI settings in %s: %s", rungms, err)

            if mpilib_root is not None:
                mpilib_path = os.path.join(mpilib_root, mpilib_subfolder)
                self.log.debug("Software root of MPI libraries set to: %s", mpilib_path)
            else:
                raise EasyBuildError("Software root of MPI libraries (%s) not found", mpilib)

        installinfo_opts["GMS_MPI_LIB"] = mpilib
        installinfo_opts["GMS_MPI_PATH"] = mpilib_path

        # Accelerators (disabled)
        # TODO: add support for GPUs
        if LooseVersion(self.version) >= LooseVersion('20230601'):
            # offloading onto Intel GPUs
            installinfo_opts["GMS_OPENMP_OFFLOAD"] = False

        # These are extra programs which for now we simply set all to FALSE
        installinfo_opts["GMS_MSUCC"] = False
        installinfo_opts["GMS_PHI"] = "none"
        installinfo_opts["GMS_SHMTYPE"] = "sysv"
        installinfo_opts["GMS_LIBCCHEM"] = False  # libcchem
        if LooseVersion(self.version) >= LooseVersion('20230601'):
            # install options for libcchem2
            installinfo_opts["GMS_HPCCHEM"] = False
            installinfo_opts["GMS_HPCCHEM_USE_DATA_SERVERS"] = False
            # build Michigan State University code
            installinfo_opts["GMS_MSUAUTO"] = False

        # Optional plug-ins and interfaces
        # libXC
        if LooseVersion(self.version) >= LooseVersion('20200101'):
            installinfo_opts['GMS_LIBXC'] = False
            if get_software_root('libxc'):
                installinfo_opts['GMS_LIBXC'] = True
                # the linker needs to be patched to use external libXC
                lixc_libs = [os.path.join(os.environ['EBROOTLIBXC'], 'lib', lib) for lib in ['libxcf03.a', 'libxc.a']]
                libxc_linker_flags = ' '.join(lixc_libs)
                try:
                    lked = os.path.join(self.builddir, 'lked')
                    for line in fileinput.input(lked, inplace=1, backup='.orig'):
                        line = re.sub(r"^(\s*set\sLIBXC_FLAGS)=.*GMS_PATH.*", r'\1="%s"' % libxc_linker_flags, line)
                        sys.stdout.write(line)
                except IOError as err:
                    raise EasyBuildError("Failed to patch %s: %s", lked, err)
        # MDI
        # needs https://github.com/MolSSI-MDI/MDI_Library
        installinfo_opts['GMS_MDI'] = False
        # NBO
        installinfo_opts['NBO'] = False
        if get_software_root('NBO'):
            installinfo_opts['NBO'] = True
        # NEO
        installinfo_opts['NEO'] = False
        # RISM
        if LooseVersion(self.version) >= LooseVersion('20230601'):
            installinfo_opts['RISM'] = False
        # TINKER
        installinfo_opts['TINKER'] = False
        if get_software_root('TINKER'):
            installinfo_opts['TINKER'] = True
        # VB2000
        installinfo_opts['VB2000'] = False
        # VM2
        installinfo_opts['GMS_VM2'] = False
        # XMVB
        installinfo_opts['XMVB'] = False

        # add include paths from dependencies
        installinfo_opts["GMS_FPE_FLAGS"] = '"%s"' % os.environ['CPPFLAGS']
        # might be useful for debugging
        # installinfo_opts["GMS_FPE_FLAGS"] = '"%s"' % os.environ['CPPFLAGS'] + "-ffpe-trap=invalid,zero,overflow"

        # write install.info file with configuration settings
        installinfo_file = os.path.join(self.builddir, GAMESS_INSTALL_INFO)
        # replace boolean options with their string representation
        boolean_opts = {opt: str(val).lower() for opt, val in installinfo_opts.items() if val in [True, False]}
        installinfo_opts.update(boolean_opts)
        # format: setenv KEY VALUE
        installinfo_txt = '\n'.join(["setenv %s %s" % (k, installinfo_opts[k]) for k in installinfo_opts])
        write_file(installinfo_file, installinfo_txt)
        self.log.debug("Contents of %s:\n%s" % (installinfo_file, read_file(installinfo_file)))

        # patch hardcoded settings in rungms to use values specified in easyconfig file
        rungms = os.path.join(self.builddir, 'rungms')
        extra_gmspath_lines = "set ERICFMT=$GMSPATH/auxdata/ericfmt.dat\nset MCPPATH=$GMSPATH/auxdata/MCP\n"
        try:
            for line in fileinput.input(rungms, inplace=1, backup='.orig'):
                line = re.sub(r"^(\s*set\s*TARGET)=.*", r"\1=%s" % self.cfg['ddi_comm'], line)
                line = re.sub(r"^(\s*set\s*GMSPATH)=.*", r"\1=%s\n%s" % (self.installdir, extra_gmspath_lines), line)
                line = re.sub(r"(null\) set VERNO)=.*", r"\1=%s" % self.version, line)
                line = re.sub(r"^(\s*set DDI_MPI_CHOICE)=.*", r"\1=%s" % mpilib, line)
                line = re.sub(r"^(\s*set DDI_MPI_ROOT)=.*%s.*" % mpilib.lower(), r"\1=%s" % mpilib_path, line)
                line = re.sub(r"^(\s*set GA_MPI_ROOT)=.*%s.*" % mpilib.lower(), r"\1=%s" % mpilib_path, line)
                # comment out all adjustments to $LD_LIBRARY_PATH that involves hardcoded paths
                line = re.sub(r"^(\s*)(setenv\s*LD_LIBRARY_PATH\s*/.*)", r"\1#\2", line)
                # scratch directory paths
                line = re.sub(r"^(\s*set\s*SCR)=.*", r"if ( ! $?SCR ) \1=%s" % self.cfg['scratch_dir'], line)
                line = re.sub(
                    r"^(\s*set\s*USERSCR)=.*", r"if ( ! $?USERSCR ) \1=%s" % self.cfg['user_scratch_dir'], line
                )
                line = re.sub(r"^(df -k \$SCR)$", r"mkdir -p $SCR && mkdir -p $USERSCR && \1", line)
                if self.cfg['hyperthreading'] is False:
                    # disable hyperthreading (1 thread per core)
                    line = re.sub(r"\$PPN \+ \$PPN", r"$PPN", line)
                    line = re.sub(r"\$NCPUS \+ \$NCPUS", r"$NCPUS", line)
                sys.stdout.write(line)
        except IOError as err:
            raise EasyBuildError("Failed to patch %s: %s", rungms, err)

        # Replacing the MAXCPUS and MAXNODES in compddi to a value from the EasyConfig file
        compddi = os.path.join(self.builddir, 'ddi/compddi')
        try:
            for line in fileinput.input(compddi, inplace=1, backup='.orig'):
                line = re.sub(r"^(\s*set MAXCPUS)=.*", r"\1=%s" % self.cfg['maxcpus'], line, 1)
                line = re.sub(r"^(\s*set MAXNODES)=.*", r"\1=%s" % self.cfg['maxnodes'], line, 1)
                sys.stdout.write(line)
        except IOError as err:
            raise EasyBuildError("Failed to patch compddi", compddi, err)

        # for GAMESS-US 20200630-R1 we need to build the actvte.x program
        if self.version == "20200630-R1":
            actvte = os.path.join(self.builddir, 'tools/actvte.code')
            try:
                for line in fileinput.input(actvte, inplace=1, backup='.orig'):
                    line = re.sub("[*]UNX", "    ", line)
                    sys.stdout.write(line)
            except IOError as err:
                raise EasyBuildError("Failed to patch actvte.code", actvte, err)
            # compiling
            run_cmd("mv %s/tools/actvte.code" % self.builddir + " %s/tools/actvte.f" % self.builddir)
            run_cmd(
                "%s -o " % fortran_comp + " %s/tools/actvte.x" % self.builddir + " %s/tools/actvte.f" % self.builddir
            )

    def build_step(self):
        """Custom build procedure for GAMESS-US: using compddi, compall and lked scripts."""
        compddi = os.path.join(self.cfg['start_dir'], 'ddi', 'compddi')
        run_cmd(compddi, log_all=True, simple=True)

        # make sure the libddi.a library is present
        libddi = os.path.join(self.cfg['start_dir'], 'ddi', 'libddi.a')
        if not os.path.isfile(libddi):
            raise EasyBuildError("The libddi.a library (%s) was never built", libddi)
        else:
            self.log.info("The libddi.a library (%s) was successfully built." % libddi)

        ddikick = os.path.join(self.cfg['start_dir'], 'ddi', 'ddikick.x')
        if os.path.isfile(ddikick):
            self.log.info("The ddikick.x executable (%s) was successfully built." % ddikick)

            if self.cfg['ddi_comm'] == 'sockets':
                src = ddikick
                dst = os.path.join(self.cfg['start_dir'], 'ddikick.x')
                self.log.info("Moving ddikick.x executable from %s to %s." % (src, dst))
                os.rename(src, dst)

        compall_cmd = os.path.join(self.cfg['start_dir'], 'compall')
        compall = "%s %s %s" % (self.cfg['prebuildopts'], compall_cmd, self.cfg['buildopts'])
        run_cmd(compall, log_all=True, simple=True)

        cmd = "%s gamess %s" % (os.path.join(self.cfg['start_dir'], 'lked'), self.version)
        run_cmd(cmd, log_all=True, simple=True)

    def test_step(self):
        """Run GAMESS-US tests (if 'runtest' easyconfig parameter is set to True)."""
        if self.cfg['runtest']:
            # Avoid provided 'runall' script for tests, since that only runs the tests in serial
            # Tests must be run in parallel for MPI builds and can be run serial for other build types

            target_tests = [
                # (test name, path to test input)
                (os.path.splitext(os.path.basename(exam_file))[0], exam_file)
                for exam_file in glob.glob(os.path.join(self.installdir, 'tests', 'standard', 'exam*.inp'))
            ]
            test_procs = "1"
            test_env_vars = ['export OMP_NUM_THREADS=1']

            if self.cfg['ddi_comm'] == 'mpi':
                if not build_option('mpi_tests'):
                    self.log.info("Skipping tests of MPI build of GAMESS-US by user request ('mpi_tests' is disabled)")
                    return

                # MPI builds can only run tests that support parallel execution
                if int(self.cfg['parallel']) < 2:
                    self.log.info("Skipping testing of GAMESS-US as MPI tests need at least 2 CPU cores to run")
                    return

                test_procs = str(self.cfg['parallel'])
                target_tests = [exam for exam in target_tests if exam[0] not in GAMESS_SERIAL_TESTS]

                if self.toolchain.mpi_family() == toolchain.INTELMPI:
                    test_env_vars.extend([
                        # enable fallback in case first fabric fails (see $I_MPI_FABRICS_LIST)
                        'export I_MPI_FALLBACK=enable',
                        # tests are only run locally (single node), so no SSH required
                        'export I_MPI_HYDRA_BOOTSTRAP=fork',
                    ])

            # Prepare test directory to run tests
            try:
                cwd = os.getcwd()
                change_dir(self.testdir)
            except OSError as err:
                raise EasyBuildError("Failed to move to temporary directory for running tests: %s", err)

            for exam, exam_file in target_tests:
                try:
                    copy_file(exam_file, self.testdir)
                except OSError as err:
                    raise EasyBuildError("Failed to copy test '%s' to %s: %s", exam, self.testdir, err)

            test_env_vars.append('SCR=%s' % self.testdir)

            # run target exam<id> tests, dump output to exam<id>.log
            rungms = os.path.join(self.installdir, 'rungms')
            for exam, exam_file in target_tests:
                rungms_prefix = ' && '.join(test_env_vars)
                test_cmd = [rungms_prefix, rungms, exam_file, self.version, test_procs, test_procs]
                (out, _) = run_cmd(' '.join(test_cmd), log_all=True, simple=False)
                write_file('%s.log' % exam, out)

            check_cmd = os.path.join(self.installdir, 'tests', 'standard', 'checktst')
            (out, _) = run_cmd(check_cmd, log_all=True, simple=False)

            # verify output of tests
            failed_regex = re.compile(r"^.*!!FAILED\.$", re.M)
            failed_tests = set([exam[0:6] for exam in failed_regex.findall(out)])
            done_tests = set([exam[0] for exam in target_tests])
            if done_tests - failed_tests == done_tests:
                info_msg = "All target tests ran successfully!"
                if self.cfg['ddi_comm'] == 'mpi':
                    info_msg += " (serial tests ignored: %s)" % ", ".join(GAMESS_SERIAL_TESTS)
                self.log.info(info_msg)
            else:
                raise EasyBuildError("ERROR: Not all target tests ran successfully")

            # cleanup
            change_dir(cwd)
            try:
                remove_dir(self.testdir)
            except OSError as err:
                raise EasyBuildError("Failed to remove test directory %s: %s", self.testdir, err)

    def install_step(self):
        """Skip install step, since we're building in the install directory."""
        pass

    def sanity_check_step(self):
        """Custom sanity check for GAMESS-US."""
        custom_paths = {
            'files': ['gamess.%s.x' % self.version, 'rungms'],
            'dirs': [],
        }
        super(EB_GAMESS_minus_US, self).sanity_check_step(custom_paths=custom_paths)

    def make_module_extra(self):
        """Define GAMESS-US specific variables in generated module file, i.e. $GAMESSUSROOT."""
        txt = super(EB_GAMESS_minus_US, self).make_module_extra()
        txt += self.module_generator.set_environment('GAMESSUSROOT', self.installdir)
        txt += self.module_generator.prepend_paths("PATH", [''])
        return txt



##
# Copyright 2009-2024 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for installing Cargo packages (Rust lang package system)

@author: Mikael Oehman (Chalmers University of Technology)
@author: Alex Domingo (Vrije Universiteit Brussel)
"""

import os
import re

import easybuild.tools.environment as env
import easybuild.tools.systemtools as systemtools
from easybuild.tools.build_log import EasyBuildError, print_warning
from easybuild.framework.easyconfig import CUSTOM
from easybuild.framework.extensioneasyblock import ExtensionEasyBlock
from easybuild.tools.filetools import extract_file, change_dir
from easybuild.tools.run import run_cmd
from easybuild.tools.config import build_option
from easybuild.tools.filetools import compute_checksum, mkdir, write_file
from easybuild.tools.toolchain.compiler import OPTARCH_GENERIC

CRATESIO_SOURCE = "https://crates.io/api/v1/crates"

CONFIG_TOML_SOURCE_VENDOR = """
[source.vendored-sources]
directory = "{vendor_dir}"

[source.crates-io]
replace-with = "vendored-sources"

"""

CONFIG_TOML_PATCH_GIT = """
[patch."{repo}"]
{crates}
"""
CONFIG_TOML_PATCH_GIT_CRATES = """{0} = {{ path = "{1}" }}
"""

CARGO_CHECKSUM_JSON = '{{"files": {{}}, "package": "{chksum}"}}'


class Cargo(ExtensionEasyBlock):
    """Support for installing Cargo packages (Rust)"""

    @staticmethod
    def extra_options(extra_vars=None):
        """Define extra easyconfig parameters specific to Cargo"""
        extra_vars = ExtensionEasyBlock.extra_options(extra_vars)
        extra_vars.update({
            'enable_tests': [True, "Enable building of tests", CUSTOM],
            'offline': [True, "Build offline", CUSTOM],
            'lto': [None, "Override default LTO flag ('fat', 'thin', 'off')", CUSTOM],
            'crates': [[], "List of (crate, version, [repo, rev]) tuples to use", CUSTOM],
        })

        return extra_vars

    @staticmethod
    def crate_src_filename(pkg_name, pkg_version, *args):
        """Crate tarball filename based on package name and version"""
        return "{0}-{1}.tar.gz".format(pkg_name, pkg_version)

    @staticmethod
    def crate_download_filename(pkg_name, pkg_version, *args):
        """Crate download filename based on package name and version"""
        return "{0}/{1}/download".format(pkg_name, pkg_version)

    def rustc_optarch(self):
        """Determines what architecture to target.
        Translates GENERIC optarch, and respects rustc specific optarch.
        General optarchs are ignored as there is no direct translation.
        """
        if systemtools.X86_64 == systemtools.get_cpu_architecture():
            generic = '-C target-cpu=x86-64'
        else:
            generic = '-C target-cpu=generic'

        optimal = '-C target-cpu=native'

        optarch = build_option('optarch')
        if optarch:
            if isinstance(optarch, dict):
                if 'rustc' in optarch:
                    rust_optarch = optarch['rustc']
                    if rust_optarch == OPTARCH_GENERIC:
                        return generic
                    else:
                        return '-' + rust_optarch
                self.log.info("no rustc information in the optarch dict, so using %s" % optimal)
            elif optarch == OPTARCH_GENERIC:
                return generic
            else:
                self.log.warning("optarch is ignored as there is no translation for rustc, so using %s" % optimal)
        return optimal

    def __init__(self, *args, **kwargs):
        """Constructor for Cargo easyblock."""
        super(Cargo, self).__init__(*args, **kwargs)
        self.cargo_home = os.path.join(self.builddir, '.cargo')
        self.vendor_dir = os.path.join(self.builddir, 'easybuild_vendor')
        env.setvar('CARGO_HOME', self.cargo_home)
        env.setvar('RUSTC', 'rustc')
        env.setvar('RUSTDOC', 'rustdoc')
        env.setvar('RUSTFMT', 'rustfmt')
        env.setvar('RUSTFLAGS', self.rustc_optarch())
        env.setvar('RUST_LOG', 'DEBUG')
        env.setvar('RUST_BACKTRACE', '1')

        # Populate sources from "crates" list of tuples
        sources = []
        for crate_info in self.crates:
            if len(crate_info) == 2:
                sources.append({
                    'download_filename': self.crate_download_filename(*crate_info),
                    'filename': self.crate_src_filename(*crate_info),
                    'source_urls': [CRATESIO_SOURCE],
                    'alt_location': 'crates.io',
                })
            else:
                crate, version, repo, rev = crate_info
                url, repo_name = repo.rsplit('/', maxsplit=1)
                if repo_name.endswith('.git'):
                    repo_name = repo_name[:-4]
                sources.append({
                    'git_config': {'url': url, 'repo_name': repo_name, 'commit': rev},
                    'filename': self.crate_src_filename(crate, version),
                })

        self.cfg.update('sources', sources)

    @property
    def crates(self):
        """Return the crates as defined in the EasyConfig"""
        return self.cfg['crates']

    def extract_step(self):
        """
        Unpack the source files and populate them with required .cargo-checksum.json if offline
        """
        mkdir(self.vendor_dir)

        vendor_crates = {self.crate_src_filename(*crate): crate for crate in self.crates}
        git_sources = {crate[2]: [] for crate in self.crates if len(crate) == 4}

        for src in self.src:
            extraction_dir = self.builddir
            # Extract dependency crates into vendor subdirectory, separate from sources of main package
            if src['name'] in vendor_crates:
                extraction_dir = self.vendor_dir

            self.log.info("Unpacking source of %s", src['name'])
            existing_dirs = set(os.listdir(extraction_dir))
            crate_dir = None
            src_dir = extract_file(src['path'], extraction_dir, cmd=src['cmd'],
                                   extra_options=self.cfg['unpack_options'], change_into_dir=False)
            new_extracted_dirs = set(os.listdir(extraction_dir)) - existing_dirs

            if len(new_extracted_dirs) == 1:
                # Expected crate tarball with 1 folder
                crate_dir = new_extracted_dirs.pop()
                src_dir = os.path.join(extraction_dir, crate_dir)
                self.log.debug("Unpacked sources of %s into: %s", src['name'], src_dir)
            elif len(new_extracted_dirs) == 0:
                # Extraction went wrong
                raise EasyBuildError("Unpacking sources of '%s' failed", src['name'])
            # TODO: properly handle case with multiple extracted folders
            # this is currently in a grey area, might still be used by cargo

            change_dir(src_dir)
            self.src[self.src.index(src)]['finalpath'] = src_dir

            if self.cfg['offline'] and crate_dir:
                # Create checksum file for extracted sources required by vendored crates.io sources
                self.log.info('creating .cargo-checksums.json file for : %s', crate_dir)
                chksum = compute_checksum(src['path'], checksum_type='sha256')
                chkfile = os.path.join(extraction_dir, crate_dir, '.cargo-checksum.json')
                write_file(chkfile, CARGO_CHECKSUM_JSON.format(chksum=chksum))
                # Add path to extracted sources for any crate from a git repo
                try:
                    crate_name, _, crate_repo, _ = vendor_crates[src['name']]
                except (ValueError, KeyError):
                    pass
                else:
                    self.log.debug("Sources of %s belong to git repo: %s", src['name'], crate_repo)
                    git_src_dir = (crate_name, src_dir)
                    git_sources[crate_repo].append(git_src_dir)

        if self.cfg['offline']:
            self.log.info("Setting vendored crates dir for offline operation")
            config_toml = os.path.join(self.cargo_home, 'config.toml')
            # Replace crates-io with vendored sources using build dir wide toml file in CARGO_HOME
            # because the rust source subdirectories might differ with python packages
            self.log.debug("Writting config.toml entry for vendored crates from crate.io")
            write_file(config_toml, CONFIG_TOML_SOURCE_VENDOR.format(vendor_dir=self.vendor_dir), append=True)

            # also vendor sources from other git sources (could be many crates for one git source)
            for git_repo, repo_crates in git_sources.items():
                self.log.debug("Writting config.toml entry for git repo: %s", git_repo)
                config_crates = ''.join([CONFIG_TOML_PATCH_GIT_CRATES.format(*crate) for crate in repo_crates])
                write_file(config_toml, CONFIG_TOML_PATCH_GIT.format(repo=git_repo, crates=config_crates), append=True)

            # Use environment variable since it would also be passed along to builds triggered via python packages
            env.setvar('CARGO_NET_OFFLINE', 'true')

    def configure_step(self):
        """Empty configuration step."""
        pass

    @property
    def profile(self):
        return 'debug' if self.toolchain.options.get('debug', None) else 'release'

    def build_step(self):
        """Build with cargo"""
        parallel = ''
        if self.cfg['parallel']:
            parallel = "-j %s" % self.cfg['parallel']

        tests = ''
        if self.cfg['enable_tests']:
            tests = "--tests"

        lto = ''
        if self.cfg['lto'] is not None:
            lto = '--config profile.%s.lto=\\"%s\\"' % (self.profile, self.cfg['lto'])

        run_cmd('rustc --print cfg', log_all=True, simple=True)  # for tracking in log file
        cmd = ' '.join([
            self.cfg['prebuildopts'],
            'cargo build',
            '--profile=' + self.profile,
            lto,
            tests,
            parallel,
            self.cfg['buildopts'],
        ])
        run_cmd(cmd, log_all=True, simple=True)

    def test_step(self):
        """Test with cargo"""
        if self.cfg['enable_tests']:
            cmd = ' '.join([
                self.cfg['pretestopts'],
                'cargo test',
                '--profile=' + self.profile,
                self.cfg['testopts'],
            ])
            run_cmd(cmd, log_all=True, simple=True)

    def install_step(self):
        """Install with cargo"""
        cmd = ' '.join([
            self.cfg['preinstallopts'],
            'cargo install',
            '--profile=' + self.profile,
            '--root=' + self.installdir,
            '--path=.',
            self.cfg['installopts'],
        ])
        run_cmd(cmd, log_all=True, simple=True)


def generate_crate_list(sourcedir):
    """Helper for generating crate list"""
    import toml

    cargo_toml = toml.load(os.path.join(sourcedir, 'Cargo.toml'))
    cargo_lock = toml.load(os.path.join(sourcedir, 'Cargo.lock'))

    try:
        app_name = cargo_toml['package']['name']
    except KeyError:
        app_name = os.path.basename(os.path.abspath(sourcedir))
        print_warning('Did not find a [package] name= entry. Assuming it is the folder name: ' + app_name)
    deps = cargo_lock['package']

    app_in_cratesio = False
    crates = []
    other_crates = []
    for dep in deps:
        name = dep['name']
        version = dep['version']
        if 'source' in dep:
            if name == app_name:
                app_in_cratesio = True  # exclude app itself, needs to be first in crates list or taken from pypi
            else:
                if dep['source'] == 'registry+https://github.com/rust-lang/crates.io-index':
                    crates.append((name, version))
                else:
                    # Lock file has #revision in the url
                    url, rev = dep['source'].rsplit('#', maxsplit=1)
                    for prefix in ('registry+', 'git+'):
                        if url.startswith(prefix):
                            url = url[len(prefix):]
                    # Remove branch name and revision URL parameters if present
                    url = re.sub(r'\?branch=\w+$', '', url)
                    url = re.sub(r'\?rev=%s+$' % rev, '', url)
                    crates.append((name, version, url, rev))
        else:
            other_crates.append((name, version))
    return app_in_cratesio, crates, other_crates


if __name__ == '__main__':
    import sys
    app_in_cratesio, crates, other = generate_crate_list(sys.argv[1])
    print(other)
    if app_in_cratesio or crates:
        print('crates = [')
        if app_in_cratesio:
            print('    (name, version),')
        for crate_info in crates:
            print("    %s," % str(crate_info))
        print(']')



##
# Copyright 2009-2024 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for installing Ansys Electronics Desktop

@author: Chia-Jung Hsu (Chalmers University of Technology)
"""
import os
import glob
import tempfile

from easybuild.easyblocks.generic.packedbinary import PackedBinary
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.run import run_cmd


class EB_AEDT(PackedBinary):
    """Support for installing Ansys Electronics Desktop."""

    def __init__(self, *args, **kwargs):
        """Initialize Ansys Electronics Desktop specific variables."""
        super(EB_AEDT, self).__init__(*args, **kwargs)
        self.subdir = None

    def _set_subdir(self):
        idirs = glob.glob(os.path.join(self.installdir, 'v*/Linux*/'))
        if len(idirs) == 1:
            self.subdir = os.path.relpath(idirs[0], self.installdir)
        else:
            raise EasyBuildError("Failed to locate single install subdirectory v*/Linux*/")

    def install_step(self):
        """Install Ansys Electronics Desktop using its setup tool."""
        licserv = os.getenv('EB_AEDT_LICENSE_SERVER')
        licport = os.getenv('EB_AEDT_LICENSE_SERVER_PORT')
        licservs = licserv.split(',')
        licservopt = ["-DLICENSE_SERVER%d=%s" % (i, serv) for i, serv in enumerate(licservs, 1)]
        redundant = len(licservs) > 1
        options = " ".join([
            "-i silent",
            "-DUSER_INSTALL_DIR=%s" % self.installdir,
            "-DTMP_DIR=%s" % tempfile.TemporaryDirectory().name,
            "-DLIBRARY_COMMON_INSTALL=1",
            "-DSPECIFY_LIC_CFG=1",
            "-DREDUNDANT_SERVERS=%d" % redundant,
            *licservopt,
            "-DSPECIFY_PORT=1",
            "-DLICENSE_PORT=%s" % licport,
        ])
        cmd = "./Linux/AnsysEM/Disk1/InstData/setup.exe %s" % options
        run_cmd(cmd, log_all=True, simple=True)

    def post_install_step(self):
        """Disable OS check and set LC_ALL/LANG for runtime"""
        if not self.subdir:
            self._set_subdir()

        # Clean script file to disable over restrict OS checking
        with open(os.path.join(self.installdir, self.subdir, "VerifyOS.bash"), "w") as f:
            f.write("")

        # Follow the settings in .setup_runtime and .setup_runtime_mpi to set LC_ALL and LANG
        # If LC_ALL is not set properly, AEDT throws the following runtime error:
        # what():  locale::facet::_S_create_c_locale name not valid
        for fname in [".setup_runtime", ".setup_runtime_mpi"]:
            with open(os.path.join(self.installdir, self.subdir, fname), "r") as f:
                orig = f.read()
            with open(os.path.join(self.installdir, self.subdir, fname), "w") as f:
                f.write("export LC_ALL=C\n")
                f.write("export LANG=C\n")
                f.write(orig)

    def make_module_extra(self):
        """Extra module entries for Ansys Electronics Desktop."""

        txt = super(EB_AEDT, self).make_module_extra()
        version = self.version[2:].replace('R', '')

        idirs = glob.glob(os.path.join(self.installdir, 'v*/Linux*/'))
        if len(idirs) == 1:
            subdir = os.path.relpath(idirs[0], self.installdir)
            # PyAEDT and other tools use the variable to find available AEDT versions
            txt += self.module_generator.set_environment('ANSYSEM_ROOT%s' % version,
                                                         os.path.join(self.installdir, self.subdir))

            txt += self.module_generator.prepend_paths('PATH', subdir)
            txt += self.module_generator.prepend_paths('LD_LIBRARY_PATH', subdir)
            txt += self.module_generator.prepend_paths('LIBRARY_PATH', subdir)

        return txt

    def sanity_check_step(self):
        """Custom sanity check for Ansys Electronics Desktop."""
        if not self.subdir:
            self._set_subdir()

        custom_paths = {
            'files': [os.path.join(self.subdir, 'ansysedt')],
            'dirs': [self.subdir],
        }

        inpfile = os.path.join(self.subdir, 'Examples', 'RMxprt', 'lssm', 'sm-1.aedt')
        custom_commands = ['ansysedt -ng -BatchSolve -Distributed %s' % inpfile]

        super(EB_AEDT, self).sanity_check_step(custom_paths=custom_paths, custom_commands=custom_commands)



##
# Copyright 2009-2024 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for Quantum ESPRESSO, implemented as an easyblock

@author: Kenneth Hoste (Ghent University)
@author: Ake Sandgren (HPC2N, Umea University)
@author: Davide Grassano (CECAM, EPFL)
"""

import fileinput
import os
import re
import shutil
import sys

import easybuild.tools.environment as env
import easybuild.tools.toolchain as toolchain
from easybuild.framework.easyblock import EasyBlock
from easybuild.framework.easyconfig import CUSTOM, EasyConfig
from easybuild.tools import LooseVersion
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.config import build_option
from easybuild.tools.filetools import copy_dir, copy_file
from easybuild.tools.modules import get_software_root, get_software_version
from easybuild.tools.run import run_cmd

from easybuild.easyblocks.generic.cmakemake import CMakeMake
from easybuild.easyblocks.generic.configuremake import ConfigureMake


class EB_QuantumESPRESSO(EasyBlock):
    @staticmethod
    def extra_options():
        """Custom easyconfig parameters for Quantum ESPRESSO."""
        extra_opts = EB_QuantumESPRESSOconfig.extra_options()
        extra_opts.update(EB_QuantumESPRESSOcmake.extra_options())

        return extra_opts

    def __getattribute__(self, name):
        try:
            ebclass = object.__getattribute__(self, 'ebclass')
        except AttributeError:
            ebclass = None
        if name == 'ebclass':
            return ebclass
        if ebclass is None:
            return object.__getattribute__(self, name)
        return ebclass.__getattribute__(name)

    def __init__(self, ec, *args, **kwargs):
        """Select the correct EB depending on version."""
        super(EB_QuantumESPRESSO, self).__init__(ec, *args, **kwargs)

        if LooseVersion(self.version) < LooseVersion('7.3.1'):
            self.log.info('Using legacy easyblock for Quantum ESPRESSO')
            eb = EB_QuantumESPRESSOconfig
        else:
            self.log.info('Using CMake easyblock for Quantum ESPRESSO')
            eb = EB_QuantumESPRESSOcmake

        # Required to avoid CMakeMake default extra_opts to override the ConfigMake ones
        new_ec = EasyConfig(ec.path, extra_options=eb.extra_options())
        self.ebclass = eb(new_ec, *args, **kwargs)

    class EB_QuantumESPRESSOcmake(CMakeMake):
        """Support for building and installing Quantum ESPRESSO."""

        TEST_SUITE_DIR = 'test-suite'
        SUBMODULES = [
            'lapack',
            'mbd',
            'devxlib',
            'fox',
            'd3q',
            'qe-gipaw',
            'pw2qmcpack',
            'wannier90'
        ]

        @staticmethod
        def extra_options():
            """Custom easyconfig parameters for Quantum ESPRESSO."""
            extra_vars = {
                'with_cuda': [False, 'Enable CUDA support', CUSTOM],
                'with_scalapack': [True, 'Enable ScaLAPACK support', CUSTOM],
                'with_fox': [False, 'Enable FoX support', CUSTOM],
                'with_gipaw': [True, 'Enable GIPAW support', CUSTOM],
                'with_d3q': [False, 'Enable D3Q support', CUSTOM],
                'with_qmcpack': [False, 'Enable QMCPACK support', CUSTOM],
                'test_suite_nprocs': [1, 'Number of processors to use for the test suite', CUSTOM],
                'test_suite_allow_failures': [
                    [],
                    'List of test suite targets that are allowed to fail (name can partially match)',
                    CUSTOM
                    ],
                'test_suite_threshold': [
                    0.97,
                    'Threshold for test suite success rate (does count also allowed failures)',
                    CUSTOM
                    ],
                'test_suite_max_failed': [
                    0,
                    'Maximum number of failing tests (does not count allowed failures)',
                    CUSTOM
                ],
            }
            return CMakeMake.extra_options(extra_vars)

        def __init__(self, *args, **kwargs):
            """Add extra config options specific to Quantum ESPRESSO."""
            super(EB_QuantumESPRESSOcmake, self).__init__(*args, **kwargs)

            self.install_subdir = 'qe-%s' % self.version

            self.check_bins = []

        def _add_toolchains_opts(self):
            """Enable toolchain options for Quantum ESPRESSO."""
            comp_fam = self.toolchain.comp_family()

            allowed_toolchains = [toolchain.INTELCOMP, toolchain.GCC, toolchain.NVHPC]
            if comp_fam not in allowed_toolchains:
                raise EasyBuildError(
                    "EasyBuild does not yet have support for QuantumESPRESSO with toolchain %s" % comp_fam
                )

            # If toolchain uses FlexiBLAS/OpenBLAS/NVHPC make sure to search for it first in cmake to avoid
            # finding system/site installed mkl libraries (QE's cmake, as of 7.3.1, tries to detect iMKL first on
            # x86_64 ystems without BLA_VENDOR set)
            # https://gitlab.com/QEF/q-e/-/blob/qe-7.3.1/CMakeLists.txt?ref_type=tags#L415
            # https://cmake.org/cmake/help/latest/module/FindBLAS.html
            # Higher level library checks first so that in a FlexiBLAS->OpenBLAS->NVHPC environment, the correct
            # library is found first
            if get_software_root('FlexiBLAS'):
                self.cfg.update('configopts', '-DBLA_VENDOR="FlexiBLAS"')
            elif get_software_root('OpenBLAS'):
                self.cfg.update('configopts', '-DBLA_VENDOR="OpenBLAS"')
            elif get_software_root('NVHPC'):
                self.cfg.update('configopts', '-DBLA_VENDOR="NVHPC"')

            self._add_mpi()
            self._add_openmp()
            self._add_cuda()

        def _add_libraries(self):
            """Enable external libraries for Quantum ESPRESSO."""
            self._add_scalapack()
            self._add_fox()
            self._add_hdf5()
            self._add_libxc()
            self._add_elpa()

        def _add_plugins(self):
            """Enable plugins for Quantum ESPRESSO."""
            plugins = []
            if self.cfg.get('with_gipaw', False):
                plugins += self._add_gipaw()
            if self.cfg.get('with_d3q', False):
                plugins += self._add_d3q()
            if self.cfg.get('with_qmcpack', False):
                plugins += self._add_qmcpack()
            if plugins:
                self.cfg.update('configopts', '-DQE_ENABLE_PLUGINS="%s"' % ';'.join(plugins))

        def _add_mpi(self):
            """Enable MPI for Quantum ESPRESSO."""
            if self.toolchain.options.get('usempi', False):
                self.cfg.update('configopts', '-DQE_ENABLE_MPI=ON')
            else:
                self.cfg.update('configopts', '-DQE_ENABLE_MPI=OFF')

        def _add_openmp(self):
            """Enable OpenMP for Quantum ESPRESSO."""
            if self.toolchain.options.get('openmp', False):
                self.cfg.update('configopts', '-DQE_ENABLE_OPENMP=ON')
            else:
                self.cfg.update('configopts', '-DQE_ENABLE_OPENMP=OFF')

        def _add_cuda(self):
            """Enable CUDA for Quantum ESPRESSO."""
            if self.cfg.get('with_cuda', False):
                self.cfg.update('configopts', '-DQE_ENABLE_CUDA=ON')
                self.cfg.update('configopts', '-DQE_ENABLE_OPENACC=ON')
            else:
                self.cfg.update('configopts', '-DQE_ENABLE_CUDA=OFF')
                self.cfg.update('configopts', '-DQE_ENABLE_OPENACC=OFF')

        def _add_scalapack(self):
            """Enable ScaLAPACK for Quantum ESPRESSO."""
            if self.cfg.get('with_scalapack', False):
                if not self.toolchain.options.get('usempi', False):
                    raise EasyBuildError('ScaLAPACK support requires MPI')
                self.cfg.update('configopts', '-DQE_ENABLE_SCALAPACK=ON')
            else:
                self.cfg.update('configopts', '-DQE_ENABLE_SCALAPACK=OFF')

        def _add_fox(self):
            """Enable FoX for Quantum ESPRESSO."""
            if self.cfg.get('with_fox', False):
                self.cfg.update('configopts', '-DQE_ENABLE_FOX=ON')
            else:
                self.cfg.update('configopts', '-DQE_ENABLE_FOX=OFF')

        def _add_hdf5(self):
            """Enable HDF5 for Quantum ESPRESSO."""
            if get_software_root('HDF5'):
                self.cfg.update('configopts', '-DQE_ENABLE_HDF5=ON')
            else:
                self.cfg.update('configopts', '-DQE_ENABLE_HDF5=OFF')

        def _add_libxc(self):
            """Enable LibXC for Quantum ESPRESSO."""
            if get_software_root('libxc'):
                self.cfg.update('configopts', '-DQE_ENABLE_LIBXC=ON')
            else:
                self.cfg.update('configopts', '-DQE_ENABLE_LIBXC=OFF')

        def _add_elpa(self):
            """Enable ELPA for Quantum ESPRESSO."""
            if get_software_root('ELPA'):
                if not self.cfg.get('with_scalapack', False):
                    raise EasyBuildError('ELPA support requires ScaLAPACK')
                if LooseVersion(self.version) == LooseVersion('7.3') and self.toolchain.options.get('openmp', False):
                    raise EasyBuildError('QE 7.3 with cmake does not support ELPA with OpenMP')
                self.cfg.update('configopts', '-DQE_ENABLE_ELPA=ON')
            else:
                self.cfg.update('configopts', '-DQE_ENABLE_ELPA=OFF')

        def _add_gipaw(self):
            """Enable GIPAW for Quantum ESPRESSO."""
            if LooseVersion(self.version) == LooseVersion('7.3.1'):
                # See issue: https://github.com/dceresoli/qe-gipaw/issues/19
                if not os.path.exists(os.path.join(self.builddir, self.install_subdir, 'external', 'qe-gipaw', '.git')):
                    raise EasyBuildError(
                        'GIPAW compilation will fail for QE 7.3.1 without submodule downloaded via' +
                        'sources in easyconfig.'
                        )
            res = ['gipaw']
            self.check_bins += ['gipaw.x']
            return res

        def _add_d3q(self):
            """Enable D3Q for Quantum ESPRESSO."""
            if LooseVersion(self.version) <= LooseVersion('7.3.1'):
                # See issues:
                # https://gitlab.com/QEF/q-e/-/issues/666
                # https://github.com/anharmonic/d3q/issues/13
                if not os.path.exists(os.path.join(self.builddir, self.install_subdir, 'external', 'd3q', '.git')):
                    raise EasyBuildError(
                        'D3Q compilation will fail for QE 7.3 and 7.3.1 without submodule downloaded via' +
                        'sources in easyconfig.'
                        )
            if not self.toolchain.options.get('usempi', False):
                raise EasyBuildError('D3Q support requires MPI enabled')
            res = ['d3q']
            self.check_bins += [
                'd3_asr3.x', 'd3_db.x', 'd3_import_shengbte.x', 'd3_interpolate2.x', 'd3_lw.x', 'd3_q2r.x',
                'd3_qha.x', 'd3_qq2rr.x', 'd3q.x', 'd3_r2q.x', 'd3_recenter.x', 'd3_rmzeu.x', 'd3_sparse.x',
                'd3_sqom.x', 'd3_tk.x',
                ]
            return res

        def _add_qmcpack(self):
            """Enable QMCPACK for Quantum ESPRESSO."""
            res = ['pw2qmcpack']
            self.check_bins += ['pw2qmcpack.x']
            return res

        def _copy_submodule_dirs(self):
            """Copy submodule dirs downloaded by EB into XXX/external"""
            for submod in self.SUBMODULES:
                src = os.path.join(self.builddir, submod)
                dst = os.path.join(self.builddir, self.install_subdir, 'external', submod)

                if os.path.exists(src):
                    self.log.info('Copying submodule %s into %s' % (submod, dst))
                    # Remove empty directories and replace them with the downloaded submodule
                    if os.path.exists(dst):
                        shutil.rmtree(dst)
                    shutil.move(src, dst)

                    # Trick QE to think that the submodule is already installed in case `keep_git_dir` is not used in
                    # the easyconfig file
                    gitf = os.path.join(dst, '.git')
                    if not os.path.exists(gitf):
                        os.mkdir(gitf)
                else:
                    self.log.warning('Submodule %s not found at %s' % (submod, src))

        def configure_step(self):
            """Custom configuration procedure for Quantum ESPRESSO."""

            if LooseVersion(self.version) < LooseVersion('7.3'):
                raise EasyBuildError('EB QuantumEspresso with cmake is implemented for versions >= 7.3')

            # Needs to be before other functions that could check existance of .git for submodules to
            # make compatibility checks
            self._copy_submodule_dirs()

            self._add_toolchains_opts()
            self._add_libraries()
            self._add_plugins()

            # Enable/configure test suite
            self._test_nprocs = self.cfg.get('test_suite_nprocs', 1)
            self.cfg.update('configopts', '-DQE_ENABLE_TEST=ON')
            self.cfg.update('configopts', '-DTESTCODE_NPROCS=%d' % self._test_nprocs)

            # Change format of timings to seconds only (from d/h/m/s)
            self.cfg.update('configopts', '-DQE_CLOCK_SECONDS=ON')

            if LooseVersion(self.version) <= LooseVersion('7.3.1'):
                # Needed to avoid a `DSO missing from command line` linking error
                # https://gitlab.com/QEF/q-e/-/issues/667
                if self.cfg.get('build_shared_libs', False):
                    ldflags = os.getenv('LDFLAGS', '')
                    ldflags += ' -Wl,--copy-dt-needed-entries '
                    env.setvar('LDFLAGS', ldflags)

            super(EB_QuantumESPRESSOcmake, self).configure_step()

        def test_step(self):
            """
            Test the compilation using Quantum ESPRESSO's test suite.
            ctest -j NCONCURRENT (NCONCURRENT = max (1, PARALLEL / NPROCS))
            """
            if not build_option('mpi_tests'):
                self.log.info("Skipping testing of QuantumESPRESSO since MPI testing is disabled")
                return

            thr = self.cfg.get('test_suite_threshold', 0.97)
            concurrent = max(1, self.cfg.get('parallel', 1) // self._test_nprocs)
            allow_fail = self.cfg.get('test_suite_allow_failures', [])

            cmd = ' '.join([
                'ctest',
                '-j%d' % concurrent,
                '--output-on-failure',
            ])

            (out, _) = run_cmd(cmd, log_all=False, log_ok=False, simple=False, regexp=False)

            # Example output:
            # 74% tests passed, 124 tests failed out of 481
            rgx = r'^ *(?P<perc>\d+)% tests passed, +(?P<failed>\d+) +tests failed out of +(?P<total>\d+)'
            mch = re.search(rgx, out, re.MULTILINE)
            if not mch:
                raise EasyBuildError('Failed to parse test suite output')

            perc = int(mch.group('perc')) / 100
            num_fail = int(mch.group('failed'))
            total = int(mch.group('total'))
            passed = total - num_fail
            failures = []  # list of tests that failed, to be logged at the end

            # Example output for reported failures:
            # 635/635 Test #570: system--epw_wfpt-correctness ......................................***Failed  3.52 sec
            self.log.debug('Test suite output:')
            self.log.debug(out)
            for line in out.splitlines():
                if '***Failed' in line:
                    for allowed in allow_fail:
                        if allowed in line:
                            self.log.info('Ignoring failure: %s' % line)
                            break
                    else:
                        failures.append(line)
                    self.log.warning(line)

            # Allow for flaky tests (eg too strict thresholds on results for structure relaxation)
            num_fail = len(failures)
            num_fail_thr = self.cfg.get('test_suite_max_failed', 0)
            self.log.info('Total tests passed %d out of %d  (%.2f%%)' % (passed, total, perc * 100))
            if failures:
                self.log.warning('The following tests failed (and are not ignored):')
                for failure in failures:
                    self.log.warning('|   ' + failure)
            if perc < thr:
                raise EasyBuildError(
                    'Test suite failed with less than %.2f %% (%.2f) success rate' % (thr * 100, perc * 100)
                    )
            if num_fail > num_fail_thr:
                raise EasyBuildError(
                    'Test suite failed with %d non-ignored failures (%d failures permitted)' % (num_fail, num_fail_thr)
                    )

            return out

        def sanity_check_step(self):
            """Custom sanity check for Quantum ESPRESSO."""

            targets = self.cfg['buildopts'].split()

            # Condition for all targets being build 'make'  or 'make all_currents'
            all_cond = len(targets) == 0 or 'all_currents' in targets
            pwall_cond = 'pwall' in targets

            # Standard binaries
            if all_cond or 'cp' in targets:
                self.check_bins += ['cp.x', 'cppp.x', 'manycp.x', 'wfdd.x']

            if all_cond or 'epw' in targets:
                self.check_bins += ['epw.x']

            if all_cond or 'gwl' in targets:
                self.check_bins += [
                    'abcoeff_to_eps.x', 'bse_main.x', 'graph.x', 'gww_fit.x', 'gww.x', 'head.x', 'memory_pw4gww.x',
                    'pw4gww.x', 'simple_bse.x', 'simple_ip.x', 'simple.x'
                    ]

            if all_cond or 'hp' in targets:
                self.check_bins += ['hp.x']

            if all_cond or 'ld1' in targets:
                self.check_bins += ['ld1.x']

            if all_cond or pwall_cond or 'neb' in targets:
                self.check_bins += ['neb.x', 'path_interpolation.x']

            if all_cond or pwall_cond or 'ph' in targets:
                self.check_bins += [
                    'alpha2f.x', 'dynmat.x', 'fd_ef.x', 'fd.x', 'lambda.x', 'phcg.x', 'postahc.x', 'q2r.x',
                    'dvscf_q2r.x', 'epa.x', 'fd_ifc.x', 'fqha.x', 'matdyn.x', 'ph.x', 'q2qstar.x'
                    ]

            if all_cond or pwall_cond or 'pp' in targets:
                self.check_bins += [
                    'average.x', 'dos_sp.x', 'ef.x', 'fermi_int_0.x', 'fermi_proj.x', 'fs.x', 'molecularpdos.x',
                    'pawplot.x', 'plotband.x', 'plotrho.x', 'ppacf.x', 'pp.x', 'pw2bgw.x', 'pw2gt.x', 'pw2wannier90.x',
                    'wannier_ham.x', 'wfck2r.x', 'bands.x', 'dos.x', 'epsilon.x', 'fermi_int_1.x', 'fermi_velocity.x',
                    'initial_state.x', 'open_grid.x', 'plan_avg.x', 'plotproj.x', 'pmw.x', 'pprism.x', 'projwfc.x',
                    'pw2critic.x', 'pw2gw.x', 'sumpdos.x', 'wannier_plot.x'
                    ]

            if all_cond or pwall_cond or 'pw' in targets:
                self.check_bins += [
                    'cell2ibrav.x', 'ev.x', 'ibrav2cell.x', 'kpoints.x', 'pwi2xsf.x', 'pw.x', 'scan_ibrav.x'
                    ]

            if all_cond or pwall_cond or 'pwcond' in targets:
                self.check_bins += ['pwcond.x']

            if all_cond or 'tddfpt' in targets:
                self.check_bins += [
                    'turbo_davidson.x', 'turbo_eels.x', 'turbo_lanczos.x', 'turbo_magnon.x', 'turbo_spectrum.x'
                    ]

            if all_cond or 'upf' in targets:
                self.check_bins += ['upfconv.x', 'virtual_v2.x']

            if all_cond or 'xspectra' in targets:
                self.check_bins += ['molecularnexafs.x', 'spectra_correction.x', 'xspectra.x']

            custom_paths = {
                'files': [os.path.join('bin', x) for x in self.check_bins],
                'dirs': []
            }

            super(EB_QuantumESPRESSOcmake, self).sanity_check_step(custom_paths=custom_paths)

    # Legacy version of Quantum ESPRESSO easyblock
    # Do not update further
    class EB_QuantumESPRESSOconfig(ConfigureMake):
        """Support for building and installing Quantum ESPRESSO."""

        TEST_SUITE_DIR = "test-suite"

        @staticmethod
        def extra_options():
            """Custom easyconfig parameters for Quantum ESPRESSO."""
            extra_vars = {
                'hybrid': [False, "Enable hybrid build (with OpenMP)", CUSTOM],
                'with_scalapack': [True, "Enable ScaLAPACK support", CUSTOM],
                'with_ace': [False, "Enable Adaptively Compressed Exchange support", CUSTOM],
                'with_fox': [False, "Enable FoX support", CUSTOM],
                'with_epw': [True, "Enable EPW support", CUSTOM],
                'with_gipaw': [True, "Enable GIPAW support", CUSTOM],
                'with_wannier90': [False, "Enable Wannier90 support", CUSTOM],
                'test_suite_targets': [[
                    "pw", "pp", "ph", "cp", "hp", "tddfpt", "epw",
                    ], "List of test suite targets to run", CUSTOM],
                'test_suite_allow_failures': [[
                    'relax',  # Too strict thresholds
                    'epw_polar',  # Too strict thresholds
                    'cp_h2o_scan_libxc',  # Too strict thresholds
                    'hp_metal_us_magn',  # Too strict thresholds
                    'hp_soc_UV_paw_magn',  # In 7.3 test has more params than the baseline
                    'ph_ahc_diam',  # Test detects a ! as an energy in baseline
                    'tddfpt_magnons_fe',  # Too strict thresholds
                ], "List of test suite targets that are allowed to fail (name can partially match)", CUSTOM],
                'test_suite_threshold': [
                    0.97,
                    "Threshold for test suite success rate (does count also allowed failures)",
                    CUSTOM
                    ],
                'test_suite_max_failed': [
                    0,
                    "Maximum number of failing tests (does not count allowed failures)",
                    CUSTOM
                ],
            }
            return ConfigureMake.extra_options(extra_vars)

        def __init__(self, *args, **kwargs):
            """Add extra config options specific to Quantum ESPRESSO."""
            super(EB_QuantumESPRESSOconfig, self).__init__(*args, **kwargs)

            self.install_subdir = "qe-%s" % self.version

        def patch_step(self):
            """Patch files from build dir (not start dir)."""
            super(EB_QuantumESPRESSOconfig, self).patch_step(beginpath=self.builddir)

        def _add_compiler_flags(self, comp_fam):
            """Add compiler flags to the build."""
            allowed_toolchains = [toolchain.INTELCOMP, toolchain.GCC]
            if comp_fam not in allowed_toolchains:
                raise EasyBuildError(
                    "EasyBuild does not yet have support for QuantumESPRESSO with toolchain %s" % comp_fam
                )

            if LooseVersion(self.version) >= LooseVersion("6.1"):
                if comp_fam == toolchain.INTELCOMP:
                    self.dflags += ["-D__INTEL_COMPILER"]
                elif comp_fam == toolchain.GCC:
                    self.dflags += ["-D__GFORTRAN__"]
            elif LooseVersion(self.version) >= LooseVersion("5.2.1"):
                if comp_fam == toolchain.INTELCOMP:
                    self.dflags += ["-D__INTEL"]
                elif comp_fam == toolchain.GCC:
                    self.dflags += ["-D__GFORTRAN"]
            elif LooseVersion(self.version) >= LooseVersion("5.0"):
                if comp_fam == toolchain.INTELCOMP:
                    self.dflags += ["-D__INTEL"]
                elif comp_fam == toolchain.GCC:
                    self.dflags += ["-D__GFORTRAN", "-D__STD_F95"]

        def _add_openmp(self):
            """Add OpenMP support to the build."""
            if self.toolchain.options.get('openmp', False) or self.cfg['hybrid']:
                self.cfg.update('configopts', '--enable-openmp')
                if LooseVersion(self.version) >= LooseVersion("6.2.1"):
                    self.dflags += ["-D_OPENMP"]
                elif LooseVersion(self.version) >= LooseVersion("5.0"):
                    self.dflags += ["-D__OPENMP"]

        def _add_mpi(self):
            """Add MPI support to the build."""
            if not self.toolchain.options.get('usempi', False):
                self.cfg.update('configopts', '--disable-parallel')
            else:
                self.cfg.update('configopts', '--enable-parallel')
                if LooseVersion(self.version) >= LooseVersion("6.0"):
                    self.dflags += ["-D__MPI"]
                elif LooseVersion(self.version) >= LooseVersion("5.0"):
                    self.dflags += ["-D__MPI", "-D__PARA"]

        def _add_scalapack(self, comp_fam):
            """Add ScaLAPACK support to the build."""
            if not self.cfg['with_scalapack']:
                self.cfg.update('configopts', '--without-scalapack')
            else:
                if comp_fam == toolchain.INTELCOMP:
                    if get_software_root("impi") and get_software_root("imkl"):
                        if LooseVersion(self.version) >= LooseVersion("6.2"):
                            self.cfg.update('configopts', '--with-scalapack=intel')
                        elif LooseVersion(self.version) >= LooseVersion("5.1.1"):
                            self.cfg.update('configopts', '--with-scalapack=intel')
                            self.repls += [
                                ('SCALAPACK_LIBS', os.getenv('LIBSCALAPACK'), False)
                            ]
                        elif LooseVersion(self.version) >= LooseVersion("5.0"):
                            self.cfg.update('configopts', '--with-scalapack=yes')
                        self.dflags += ["-D__SCALAPACK"]
                elif comp_fam == toolchain.GCC:
                    if get_software_root("OpenMPI") and get_software_root("ScaLAPACK"):
                        self.cfg.update('configopts', '--with-scalapack=yes')
                        self.dflags += ["-D__SCALAPACK"]
                else:
                    self.cfg.update('configopts', '--without-scalapack')

        def _add_libxc(self):
            """Add libxc support to the build."""
            libxc = get_software_root("libxc")
            if libxc:
                libxc_v = get_software_version("libxc")
                if LooseVersion(libxc_v) < LooseVersion("3.0.1"):
                    raise EasyBuildError("Must use libxc >= 3.0.1")
                if LooseVersion(self.version) >= LooseVersion("7.0"):
                    if LooseVersion(libxc_v) < LooseVersion("4"):
                        raise EasyBuildError("libxc support for QuantumESPRESSO 7.x only available for libxc >= 4")
                    self.cfg.update('configopts', '--with-libxc=yes')
                    self.cfg.update('configopts', '--with-libxc-prefix=%s' % libxc)
                elif LooseVersion(self.version) >= LooseVersion("6.6"):
                    if LooseVersion(libxc_v) >= LooseVersion("6.0"):
                        raise EasyBuildError(
                            "libxc support for QuantumESPRESSO 6.6 to 6.8 only available for libxc < 6.0"
                        )
                    if LooseVersion(libxc_v) < LooseVersion("4"):
                        raise EasyBuildError("libxc support for QuantumESPRESSO 6.x only available for libxc >= 4")
                    self.cfg.update('configopts', '--with-libxc=yes')
                    self.cfg.update('configopts', '--with-libxc-prefix=%s' % libxc)
                elif LooseVersion(self.version) >= LooseVersion("6.0"):
                    if LooseVersion(libxc_v) >= LooseVersion("5.0"):
                        raise EasyBuildError(
                            "libxc support for QuantumESPRESSO 6.0 to 6.5 only available for libxc <= 4.3.4"
                            )
                    if LooseVersion(libxc_v) < LooseVersion("4"):
                        raise EasyBuildError("libxc support for QuantumESPRESSO 6.x only available for libxc >= 4")
                    self.cfg.update('configopts', '--with-libxc=yes')
                    self.cfg.update('configopts', '--with-libxc-prefix=%s' % libxc)
                else:
                    self.extra_libs += ['-L%s/lib' % libxc, '-lxcf90', '-lxc']

                self.dflags += ["-D__LIBXC"]

        def _add_hdf5(self):
            """Add HDF5 support to the build."""
            hdf5 = get_software_root("HDF5")
            if hdf5:
                self.cfg.update('configopts', '--with-hdf5=%s' % hdf5)
                self.dflags += ["-D__HDF5"]
                hdf5_lib_repl = '-L%s/lib -lhdf5hl_fortran -lhdf5_hl -lhdf5_fortran -lhdf5 -lsz -lz -ldl -lm' % hdf5
                self.repls += [('HDF5_LIB', hdf5_lib_repl, False)]

                if LooseVersion(self.version) >= LooseVersion("6.2.1"):
                    pass
                else:
                    # Should be experimental in 6.0 but gives segfaults when used
                    raise EasyBuildError("HDF5 support is only available in QuantumESPRESSO 6.2.1 and later")

        def _add_elpa(self):
            """Add ELPA support to the build."""
            elpa = get_software_root("ELPA")
            if elpa:
                elpa_v = get_software_version("ELPA")

                if LooseVersion(elpa_v) < LooseVersion("2015"):
                    raise EasyBuildError("ELPA versions lower than 2015 are not supported")

                flag = True
                if LooseVersion(self.version) >= LooseVersion("6.8"):
                    if LooseVersion(elpa_v) >= LooseVersion("2018.11"):
                        self.dflags += ["-D__ELPA"]
                    elif LooseVersion(elpa_v) >= LooseVersion("2016.11"):
                        self.dflags += ["-D__ELPA_2016"]
                    elif LooseVersion(elpa_v) >= LooseVersion("2015"):
                        self.dflags += ["-D__ELPA_2015"]
                elif LooseVersion(self.version) >= LooseVersion("6.6"):
                    if LooseVersion(elpa_v) >= LooseVersion("2020"):
                        raise EasyBuildError("ELPA support for QuantumESPRESSO 6.6/6.7 only available up to v2019.xx")
                    elif LooseVersion(elpa_v) >= LooseVersion("2018"):
                        self.dflags += ["-D__ELPA"]
                    elif LooseVersion(elpa_v) >= LooseVersion("2015"):
                        elpa_year_v = elpa_v.split('.')[0]
                        self.dflags += ["-D__ELPA_%s" % elpa_year_v]
                elif LooseVersion(self.version) >= LooseVersion("6.0"):
                    if LooseVersion(elpa_v) >= LooseVersion("2017"):
                        raise EasyBuildError("ELPA support for QuantumESPRESSO 6.x only available up to v2016.xx")
                    elif LooseVersion(elpa_v) >= LooseVersion("2016"):
                        self.dflags += ["-D__ELPA_2016"]
                    elif LooseVersion(elpa_v) >= LooseVersion("2015"):
                        self.dflags += ["-D__ELPA_2015"]
                elif LooseVersion(self.version) >= LooseVersion("5.4"):
                    self.dflags += ["-D__ELPA"]
                    self.cfg.update('configopts', '--with-elpa=%s' % elpa)
                    flag = False
                elif LooseVersion(self.version) >= LooseVersion("5.1.1"):
                    self.cfg.update('configopts', '--with-elpa=%s' % elpa)
                    flag = False
                else:
                    raise EasyBuildError("ELPA support is only available in QuantumESPRESSO 5.1.1 and later")

                if flag:
                    if self.toolchain.options.get('openmp', False):
                        elpa_include = 'elpa_openmp-%s' % elpa_v
                        elpa_lib = 'libelpa_openmp.a'
                    else:
                        elpa_include = 'elpa-%s' % elpa_v
                        elpa_lib = 'libelpa.a'
                    elpa_include = os.path.join(elpa, 'include', elpa_include, 'modules')
                    elpa_lib = os.path.join(elpa, 'lib', elpa_lib)
                    self.repls += [
                        ('IFLAGS', '-I%s' % elpa_include, True)
                        ]
                    self.cfg.update('configopts', '--with-elpa-include=%s' % elpa_include)
                    self.cfg.update('configopts', '--with-elpa-lib=%s' % elpa_lib)
                    if LooseVersion(self.version) < LooseVersion("7.0"):
                        self.repls += [
                            ('SCALAPACK_LIBS', '%s %s' % (elpa_lib, os.getenv("LIBSCALAPACK")), False)
                            ]

        def _add_fftw(self, comp_fam):
            """Add FFTW support to the build."""
            if self.toolchain.options.get('openmp', False):
                libfft = os.getenv('LIBFFT_MT')
            else:
                libfft = os.getenv('LIBFFT')

            if LooseVersion(self.version) >= LooseVersion("5.2.1"):
                if comp_fam == toolchain.INTELCOMP and get_software_root("imkl"):
                    self.dflags += ["-D__DFTI"]
                elif libfft:
                    self.dflags += ["-D__FFTW"] if "fftw3" not in libfft else ["-D__FFTW3"]
                    self.repls += [
                        ('FFT_LIBS', libfft, False),
                    ]
            elif LooseVersion(self.version) >= LooseVersion("5.0"):
                if libfft:
                    self.dflags += ["-D__FFTW"] if "fftw3" not in libfft else ["-D__FFTW3"]
                    self.repls += [
                        ('FFT_LIBS', libfft, False),
                    ]

        def _add_ace(self):
            """Add ACE support to the build."""
            if self.cfg['with_ace']:
                if LooseVersion(self.version) >= LooseVersion("6.2"):
                    self.log.warning("ACE support is not available in QuantumESPRESSO >= 6.2")
                elif LooseVersion(self.version) >= LooseVersion("6.0"):
                    self.dflags += ["-D__EXX_ACE"]
                else:
                    self.log.warning("ACE support is not available in QuantumESPRESSO < 6.0")

        def _add_beef(self):
            """Add BEEF support to the build."""
            if LooseVersion(self.version) == LooseVersion("6.6"):
                libbeef = get_software_root("libbeef")
                if libbeef:
                    self.dflags += ["-Duse_beef"]
                    libbeef_lib = os.path.join(libbeef, 'lib')
                    self.cfg.update('configopts', '--with-libbeef-prefix=%s' % libbeef_lib)
                    self.repls += [
                        ('BEEF_LIBS_SWITCH', 'external', False),
                        ('BEEF_LIBS', str(os.path.join(libbeef_lib, "libbeef.a")), False)
                    ]

        def _add_fox(self):
            """Add FoX support to the build."""
            if self.cfg['with_fox']:
                if LooseVersion(self.version) >= LooseVersion("7.2"):
                    self.cfg.update('configopts', '--with-fox=yes')

        def _add_epw(self):
            """Add EPW support to the build."""
            if self.cfg['with_epw']:
                if LooseVersion(self.version) >= LooseVersion("6.0"):
                    self.cfg.update('buildopts', 'epw', allow_duplicate=False)
                    self.cfg.update('test_suite_targets', ['epw'], allow_duplicate=False)
                else:
                    self.log.warning("EPW support is not available in QuantumESPRESSO < 6.0")
            else:
                if 'epw' in self.cfg['buildopts']:
                    self.cfg['buildopts'] = self.cfg['buildopts'].replace('epw', '')
                if 'epw' in self.cfg['test_suite_targets']:
                    self.cfg['test_suite_targets'].remove('epw')

        def _add_gipaw(self):
            """Add GIPAW support to the build."""
            if self.cfg['with_gipaw']:
                self.cfg.update('buildopts', 'gipaw', allow_duplicate=False)
            else:
                if 'gipaw' in self.cfg['buildopts']:
                    self.cfg['buildopts'] = self.cfg['buildopts'].replace('gipaw', '')

        def _add_wannier90(self):
            """Add Wannier90 support to the build."""
            if self.cfg['with_wannier90']:
                self.cfg.update('buildopts', 'w90', allow_duplicate=False)
            else:
                if 'w90' in self.cfg['buildopts']:
                    self.cfg['buildopts'] = self.cfg['buildopts'].replace('w90', '')

        def _adjust_compiler_flags(self, comp_fam):
            """Adjust compiler flags based on the compiler family and code version."""
            if comp_fam == toolchain.INTELCOMP:
                if LooseVersion("6.0") <= LooseVersion(self.version) <= LooseVersion("6.4"):
                    i_mpi_cc = os.getenv('I_MPI_CC', '')
                    if i_mpi_cc == 'icx':
                        env.setvar('I_MPI_CC', 'icc')  # Needed as clib/qmmm_aux.c using <math.h> implicitly
            elif comp_fam == toolchain.GCC:
                pass

        def configure_step(self):
            """Custom configuration procedure for Quantum ESPRESSO."""

            if LooseVersion(self.version) >= LooseVersion("7.3.1"):
                raise EasyBuildError(
                    "QuantumESPRESSO 7.3.1 and later are not supported with the this easyblock (ConfigureMake), " +
                    "use the EB_QuantumESPRESSOcmake (CMakeMake) easyblock instead."
                    )

            # compose list of DFLAGS (flag, value, keep_stuff)
            # for guidelines, see include/defs.h.README in sources
            self.dflags = []
            self.repls = []
            self.extra_libs = []

            comp_fam = self.toolchain.comp_family()

            self._add_compiler_flags(comp_fam)
            self._add_openmp()
            self._add_mpi()
            self._add_scalapack(comp_fam)
            self._add_libxc()
            self._add_hdf5()
            self._add_elpa()
            self._add_fftw(comp_fam)
            self._add_ace()
            self._add_beef()
            self._add_fox()
            self._add_epw()
            self._add_gipaw()
            self._add_wannier90()

            if comp_fam == toolchain.INTELCOMP:
                # Intel compiler must have -assume byterecl (see install/configure)
                self.repls.append(('F90FLAGS', '-fpp -assume byterecl', True))
                self.repls.append(('FFLAGS', '-assume byterecl', True))
            elif comp_fam == toolchain.GCC:
                f90_flags = ['-cpp']
                if LooseVersion(get_software_version('GCC')) >= LooseVersion('10'):
                    f90_flags.append('-fallow-argument-mismatch')
                self.repls.append(('F90FLAGS', ' '.join(f90_flags), True))

            self._adjust_compiler_flags(comp_fam)

            super(EB_QuantumESPRESSOconfig, self).configure_step()

            # always include -w to supress warnings
            self.dflags.append('-w')

            self.repls.append(('DFLAGS', ' '.join(self.dflags), False))

            # complete C/Fortran compiler and LD flags
            if self.toolchain.options.get('openmp', False) or self.cfg['hybrid']:
                self.repls.append(('LDFLAGS', self.toolchain.get_flag('openmp'), True))
                self.repls.append(('(?:C|F90|F)FLAGS', self.toolchain.get_flag('openmp'), True))

            # libs is being used for the replacement in the wannier90 files
            libs = []
            # Only overriding for gcc as the intel flags are already being properly
            # set.
            if comp_fam == toolchain.GCC:
                num_libs = ['BLAS', 'LAPACK', 'FFT']
                if self.cfg['with_scalapack']:
                    num_libs.extend(['SCALAPACK'])
                elpa = get_software_root('ELPA')
                elpa_lib = 'libelpa_openmp.a' if self.toolchain.options.get('openmp', False) else 'libelpa.a'
                elpa_lib = os.path.join(elpa or '', 'lib', elpa_lib)
                for lib in num_libs:
                    if self.toolchain.options.get('openmp', False):
                        val = os.getenv('LIB%s_MT' % lib)
                    else:
                        val = os.getenv('LIB%s' % lib)
                    if lib == 'SCALAPACK' and elpa:
                        val = ' '.join([elpa_lib, val])
                    self.repls.append(('%s_LIBS' % lib, val, False))
                    libs.append(val)
            libs = ' '.join(libs)

            self.repls.append(('BLAS_LIBS_SWITCH', 'external', False))
            self.repls.append(('LAPACK_LIBS_SWITCH', 'external', False))
            self.repls.append(('LD_LIBS', ' '.join(self.extra_libs + [os.getenv('LIBS')]), False))

            # Do not use external FoX.
            # FoX starts to be used in 6.2 and they use a patched version that
            # is newer than FoX 4.1.2 which is the latest release.
            # Ake Sandgren, 20180712
            if get_software_root('FoX'):
                raise EasyBuildError(
                    "Found FoX external module, QuantumESPRESSO must use the version they include with the source."
                )

            self.log.info("List of replacements to perform: %s" % str(self.repls))

            if LooseVersion(self.version) >= LooseVersion("6"):
                make_ext = '.inc'
            else:
                make_ext = '.sys'

            # patch make.sys file
            fn = os.path.join(self.cfg['start_dir'], 'make' + make_ext)
            try:
                for line in fileinput.input(fn, inplace=1, backup='.orig.eb'):
                    for (k, v, keep) in self.repls:
                        # need to use [ \t]* instead of \s*, because vars may be undefined as empty,
                        # and we don't want to include newlines
                        if keep:
                            line = re.sub(r"^(%s\s*=[ \t]*)(.*)$" % k, r"\1\2 %s" % v, line)
                        else:
                            line = re.sub(r"^(%s\s*=[ \t]*).*$" % k, r"\1%s" % v, line)

                    # fix preprocessing directives for .f90 files in make.sys if required
                    if LooseVersion(self.version) < LooseVersion("6.0"):
                        if comp_fam == toolchain.GCC:
                            line = re.sub(
                                r"^\t\$\(MPIF90\) \$\(F90FLAGS\) -c \$<",
                                "\t$(CPP) -C $(CPPFLAGS) $< -o $*.F90\n" +
                                "\t$(MPIF90) $(F90FLAGS) -c $*.F90 -o $*.o",
                                line
                            )

                    if LooseVersion(self.version) == LooseVersion("6.6"):
                        # fix order of BEEF_LIBS in QE_LIBS
                        line = re.sub(
                            r"^(QELIBS\s*=[ \t]*)(.*) \$\(BEEF_LIBS\) (.*)$",
                            r"QELIBS = $(BEEF_LIBS) \2 \3", line
                            )

                        # use FCCPP instead of CPP for Fortran headers
                        line = re.sub(
                            r"\t\$\(CPP\) \$\(CPPFLAGS\) \$< -o \$\*\.fh",
                            "\t$(FCCPP) $(CPPFLAGS) $< -o $*.fh",
                            line
                            )

                    sys.stdout.write(line)
            except IOError as err:
                raise EasyBuildError("Failed to patch %s: %s", fn, err)

            with open(fn, "r") as f:
                self.log.info("Contents of patched %s: %s" % (fn, f.read()))

            # patch default make.sys for wannier
            if LooseVersion(self.version) >= LooseVersion("5"):
                fn = os.path.join(self.cfg['start_dir'], 'install', 'make_wannier90' + make_ext)
            else:
                fn = os.path.join(self.cfg['start_dir'], 'plugins', 'install', 'make_wannier90.sys')
            try:
                for line in fileinput.input(fn, inplace=1, backup='.orig.eb'):
                    if libs:
                        line = re.sub(r"^(LIBS\s*=\s*).*", r"\1%s" % libs, line)

                    sys.stdout.write(line)

            except IOError as err:
                raise EasyBuildError("Failed to patch %s: %s", fn, err)

            with open(fn, "r") as f:
                self.log.info("Contents of patched %s: %s" % (fn, f.read()))

            # patch Makefile of want plugin
            wantprefix = 'want-'
            wantdirs = [d for d in os.listdir(self.builddir) if d.startswith(wantprefix)]

            if len(wantdirs) > 1:
                raise EasyBuildError("Found more than one directory with %s prefix, help!", wantprefix)

            if len(wantdirs) != 0:
                wantdir = os.path.join(self.builddir, wantdirs[0])
                make_sys_in_path = None
                cand_paths = [os.path.join('conf', 'make.sys.in'), os.path.join('config', 'make.sys.in')]
                for path in cand_paths:
                    full_path = os.path.join(wantdir, path)
                    if os.path.exists(full_path):
                        make_sys_in_path = full_path
                        break
                if make_sys_in_path is None:
                    raise EasyBuildError(
                        "Failed to find make.sys.in in want directory %s, paths considered: %s",
                        wantdir, ', '.join(cand_paths)
                    )

                try:
                    for line in fileinput.input(make_sys_in_path, inplace=1, backup='.orig.eb'):
                        # fix preprocessing directives for .f90 files in make.sys if required
                        if comp_fam == toolchain.GCC:
                            line = re.sub(
                                "@f90rule@",
                                "$(CPP) -C $(CPPFLAGS) $< -o $*.F90\n" +
                                "\t$(MPIF90) $(F90FLAGS) -c $*.F90 -o $*.o",
                                line
                            )

                        sys.stdout.write(line)
                except IOError as err:
                    raise EasyBuildError("Failed to patch %s: %s", fn, err)

            # move non-espresso directories to where they're expected and create symlinks
            try:
                dirnames = [d for d in os.listdir(self.builddir) if d not in [self.install_subdir, 'd3q-latest']]
                targetdir = os.path.join(self.builddir, self.install_subdir)
                for dirname in dirnames:
                    shutil.move(os.path.join(self.builddir, dirname), os.path.join(targetdir, dirname))
                    self.log.info("Moved %s into %s" % (dirname, targetdir))

                    dirname_head = dirname.split('-')[0]
                    # Handle the case where the directory is preceded by 'qe-'
                    if dirname_head == 'qe':
                        dirname_head = dirname.split('-')[1]
                    linkname = None
                    if dirname_head == 'sax':
                        linkname = 'SaX'
                    if dirname_head == 'wannier90':
                        linkname = 'W90'
                    elif dirname_head in ['d3q', 'gipaw', 'plumed', 'want', 'yambo']:
                        linkname = dirname_head.upper()
                    if linkname:
                        os.symlink(os.path.join(targetdir, dirname), os.path.join(targetdir, linkname))

            except OSError as err:
                raise EasyBuildError("Failed to move non-espresso directories: %s", err)

        def test_step(self):
            """
            Test the compilation using Quantum ESPRESSO's test suite.
            cd test-suite && make run-tests NPROCS=XXX (XXX <= 4)
            """
            if not build_option('mpi_tests'):
                self.log.info("Skipping testing of QuantumESPRESSO since MPI testing is disabled")
                return

            thr = self.cfg.get('test_suite_threshold', 0.9)
            stot = 0
            spass = 0
            parallel = min(4, self.cfg.get('parallel', 1))
            test_dir = os.path.join(self.start_dir, self.TEST_SUITE_DIR)

            pseudo_loc = "https://pseudopotentials.quantum-espresso.org/upf_files/"
            # NETWORK_PSEUDO in test_suite/ENVIRONMENT is set to old url for qe 7.0 and older
            if LooseVersion(self.version) < LooseVersion("7.1"):
                cmd = ' && '.join([
                    "cd %s" % test_dir,
                    "sed -i 's|export NETWORK_PSEUDO=.*|export NETWORK_PSEUDO=%s|g' ENVIRONMENT" % pseudo_loc
                ])
                run_cmd(cmd, log_all=False, log_ok=False, simple=False, regexp=False)

            targets = self.cfg.get('test_suite_targets', [])
            allow_fail = self.cfg.get('test_suite_allow_failures', [])

            full_out = ''
            failures = []
            for target in targets:
                pcmd = ''
                if LooseVersion(self.version) < LooseVersion("7.2"):
                    if parallel > 1:
                        target = target + "-parallel"
                    else:
                        target = target + "-serial"
                else:
                    pcmd = 'NPROCS=%d' % parallel

                cmd = 'cd %s && %s make run-tests-%s' % (test_dir, pcmd, target)
                (out, _) = run_cmd(cmd, log_all=False, log_ok=False, simple=False, regexp=False)

                # Example output:
                # All done. 2 out of 2 tests passed.
                # All done. ERROR: only 6 out of 9 tests passed
                _tot = 0
                _pass = 0
                rgx = r'All done. (ERROR: only )?(?P<succeeded>\d+) out of (?P<total>\d+) tests passed.'
                for mch in re.finditer(rgx, out):
                    succeeded = int(mch.group('succeeded'))
                    total = int(mch.group('total'))
                    _tot += total
                    _pass += succeeded

                perc = _pass / max(_tot, 1)
                self.log.info("%s: Passed %d out of %d  (%.2f%%)" % (target, _pass, _tot, perc * 100))

                # Log test-suite errors if present
                if _pass < _tot:
                    # Example output for reported failures:
                    # pw_plugins - plugin-pw2casino_1.in (arg(s): 1): **FAILED**.
                    # Different sets of data extracted from benchmark and test.
                    #     Data only in benchmark: p1.
                    # (empty line)
                    flag = False
                    for line in out.splitlines():
                        if '**FAILED**' in line:
                            for allowed in allow_fail:
                                if allowed in line:
                                    self.log.info('Ignoring failure: %s' % line)
                                    break
                            else:
                                failures.append(line)
                            flag = True
                            self.log.warning(line)
                            continue
                        elif line.strip() == '':
                            flag = False
                        if flag:
                            self.log.warning('|   ' + line)

                stot += _tot
                spass += _pass
                full_out += out

            # Allow for flaky tests (eg too strict thresholds on results for structure relaxation)
            num_fail = len(failures)
            num_fail_thr = self.cfg.get('test_suite_max_failed', 0)
            perc = spass / max(stot, 1)
            self.log.info("Total tests passed %d out of %d  (%.2f%%)" % (spass, stot, perc * 100))
            if failures:
                self.log.warning("The following tests failed:")
                for failure in failures:
                    self.log.warning('|   ' + failure)
            if perc < thr:
                raise EasyBuildError(
                    "Test suite failed with less than %.2f %% (%.2f) success rate" % (thr * 100, perc * 100)
                    )
            if num_fail > num_fail_thr:
                raise EasyBuildError(
                    "Test suite failed with %d failures (%d failures permitted)" % (num_fail, num_fail_thr)
                    )

            return full_out

        def install_step(self):
            """Custom install step for Quantum ESPRESSO."""

            # In QE 7.3 the w90 target is always invoked (even if only used as a library), and the symlink to the
            # `wannier90.x` executable is generated, but the actual binary is not built. We need to remove the symlink
            if LooseVersion(self.version) == LooseVersion("7.3"):
                w90_path = os.path.join(self.start_dir, 'bin', 'wannier90.x')
                if os.path.islink(w90_path) and not os.path.exists(os.readlink(w90_path)):
                    os.unlink(w90_path)

            # extract build targets as list
            targets = self.cfg['buildopts'].split()

            # Copy all binaries
            bindir = os.path.join(self.installdir, 'bin')
            copy_dir(os.path.join(self.cfg['start_dir'], 'bin'), bindir)

            # Pick up files not installed in bin
            def copy_binaries(path):
                full_dir = os.path.join(self.cfg['start_dir'], path)
                self.log.info("Looking for binaries in %s" % full_dir)
                for filename in os.listdir(full_dir):
                    full_path = os.path.join(full_dir, filename)
                    if os.path.isfile(full_path):
                        if filename.endswith('.x'):
                            copy_file(full_path, bindir)

            if 'upf' in targets or 'all' in targets:
                if LooseVersion(self.version) < LooseVersion("6.6"):
                    copy_binaries('upftools')
                else:
                    copy_binaries('upflib')
                    copy_file(os.path.join(self.cfg['start_dir'], 'upflib', 'fixfiles.py'), bindir)

            if 'want' in targets:
                copy_binaries('WANT')

            if 'w90' in targets:
                copy_binaries('W90')

            if 'yambo' in targets:
                copy_binaries('YAMBO')

        def sanity_check_step(self):
            """Custom sanity check for Quantum ESPRESSO."""

            # extract build targets as list
            targets = self.cfg['buildopts'].split()

            bins = []
            if LooseVersion(self.version) < LooseVersion("6.7"):
                # build list of expected binaries based on make targets
                bins.extend(["iotk", "iotk.x", "iotk_print_kinds.x"])

            if 'cp' in targets or 'all' in targets:
                bins.extend(["cp.x", "wfdd.x"])
                if LooseVersion(self.version) < LooseVersion("6.4"):
                    bins.append("cppp.x")

            # only for v4.x, not in v5.0 anymore, called gwl in 6.1 at least
            if 'gww' in targets or 'gwl' in targets:
                bins.extend(["gww_fit.x", "gww.x", "head.x", "pw4gww.x"])

            if 'ld1' in targets or 'all' in targets:
                bins.extend(["ld1.x"])

            if 'gipaw' in targets:
                bins.extend(["gipaw.x"])

            if 'neb' in targets or 'pwall' in targets or 'all' in targets:
                if LooseVersion(self.version) > LooseVersion("5"):
                    bins.extend(["neb.x", "path_interpolation.x"])

            if 'ph' in targets or 'all' in targets:
                bins.extend(["dynmat.x", "lambda.x", "matdyn.x", "ph.x", "phcg.x", "q2r.x"])
                if LooseVersion(self.version) < LooseVersion("6"):
                    bins.extend(["d3.x"])
                if LooseVersion(self.version) > LooseVersion("5"):
                    bins.extend(["fqha.x", "q2qstar.x"])

            if 'pp' in targets or 'pwall' in targets or 'all' in targets:
                bins.extend([
                    "average.x", "bands.x", "dos.x", "epsilon.x", "initial_state.x", "plan_avg.x", "plotband.x",
                    "plotproj.x", "plotrho.x", "pmw.x", "pp.x", "projwfc.x", "sumpdos.x", "pw2wannier90.x", "pw2gw.x",
                    "wannier_ham.x", "wannier_plot.x"
                ])
                if LooseVersion(self.version) > LooseVersion("5") and LooseVersion(self.version) < LooseVersion("6.4"):
                    bins.extend(["pw2bgw.x", "bgw2pw.x"])
                elif LooseVersion(self.version) <= LooseVersion("5"):
                    bins.extend(["pw2casino.x"])
                if LooseVersion(self.version) < LooseVersion("6.4"):
                    bins.extend(["pw_export.x"])

            if 'pw' in targets or 'all' in targets:
                bins.extend(["dist.x", "ev.x", "kpoints.x", "pw.x", "pwi2xsf.x"])
                if LooseVersion(self.version) < LooseVersion("6.5"):
                    if LooseVersion(self.version) >= LooseVersion("5.1"):
                        bins.extend(["generate_rVV10_kernel_table.x"])
                    if LooseVersion(self.version) > LooseVersion("5"):
                        bins.extend(["generate_vdW_kernel_table.x"])
                if LooseVersion(self.version) <= LooseVersion("5"):
                    bins.extend(["path_int.x"])
                if LooseVersion(self.version) < LooseVersion("5.3"):
                    bins.extend(["band_plot.x", "bands_FS.x", "kvecs_FS.x"])

            if 'pwcond' in targets or 'pwall' in targets or 'all' in targets:
                bins.extend(["pwcond.x"])

            if 'tddfpt' in targets or 'all' in targets:
                if LooseVersion(self.version) > LooseVersion("5"):
                    bins.extend(["turbo_lanczos.x", "turbo_spectrum.x"])

            upftools = []
            if 'upf' in targets or 'all' in targets:
                if LooseVersion(self.version) < LooseVersion("6.6"):
                    upftools = ["casino2upf.x", "cpmd2upf.x", "fhi2upf.x", "fpmd2upf.x", "ncpp2upf.x",
                                "oldcp2upf.x", "read_upf_tofile.x", "rrkj2upf.x", "uspp2upf.x", "vdb2upf.x"]
                    if LooseVersion(self.version) > LooseVersion("5"):
                        upftools.extend(["interpolate.x", "upf2casino.x"])
                    if LooseVersion(self.version) >= LooseVersion("6.3"):
                        upftools.extend(["fix_upf.x"])
                    if LooseVersion(self.version) < LooseVersion("6.4"):
                        upftools.extend(["virtual.x"])
                    else:
                        upftools.extend(["virtual_v2.x"])
                else:
                    upftools = ["upfconv.x", "virtual_v2.x", "fixfiles.py"]

            if 'vdw' in targets:  # only for v4.x, not in v5.0 anymore
                bins.extend(["vdw.x"])

            if 'w90' in targets:
                bins.extend(["wannier90.x"])
                if LooseVersion(self.version) >= LooseVersion("5.4"):
                    bins.extend(["postw90.x"])
                    if LooseVersion(self.version) < LooseVersion("6.1"):
                        bins.extend(["w90chk2chk.x"])

            want_bins = []
            if 'want' in targets:
                want_bins = [
                    "blc2wan.x", "conductor.x", "current.x", "disentangle.x", "dos.x", "gcube2plt.x", "kgrid.x",
                    "midpoint.x", "plot.x", "sumpdos", "wannier.x", "wfk2etsf.x"
                ]
                if LooseVersion(self.version) > LooseVersion("5"):
                    want_bins.extend(["cmplx_bands.x", "decay.x", "sax2qexml.x", "sum_sgm.x"])

            if 'xspectra' in targets:
                bins.extend(["xspectra.x"])

            yambo_bins = []
            if 'yambo' in targets:
                yambo_bins = ["a2y", "p2y", "yambo", "ypp"]

            d3q_bins = []
            if 'd3q' in targets:
                d3q_bins = [
                    'd3_asr3.x', 'd3_lw.x', 'd3_q2r.x', 'd3_qq2rr.x', 'd3q.x', 'd3_r2q.x', 'd3_recenter.x',
                    'd3_sparse.x', 'd3_sqom.x', 'd3_tk.x'
                ]
                if LooseVersion(self.version) < LooseVersion("6.4"):
                    d3q_bins.append('d3_import3py.x')

            custom_paths = {
                'files': [os.path.join('bin', x) for x in bins + upftools + want_bins + yambo_bins + d3q_bins],
                'dirs': []
            }

            super(EB_QuantumESPRESSOconfig, self).sanity_check_step(custom_paths=custom_paths)


EB_QuantumESPRESSOconfig = EB_QuantumESPRESSO.EB_QuantumESPRESSOconfig
EB_QuantumESPRESSOcmake = EB_QuantumESPRESSO.EB_QuantumESPRESSOcmake



##
# Copyright 2013-2024 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for building and installing DualSPHysics, implemented as an easyblock

@author: Jasper Grimm (University of York)
"""
import glob
import os
import stat

from easybuild.easyblocks.generic.cmakemakecp import CMakeMakeCp
from easybuild.framework.easyconfig import CUSTOM
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.config import build_option
from easybuild.tools.filetools import adjust_permissions
from easybuild.tools.modules import get_software_root
from easybuild.tools.run import run_cmd


class EB_DualSPHysics(CMakeMakeCp):
    """Support for building/installing DualSPHysics."""

    @staticmethod
    def extra_options():
        """Extra easyconfig parameters for DualSPHysics."""
        extra_vars = CMakeMakeCp.extra_options()

        extra_vars['separate_build_dir'][0] = True

        # files_to_copy is not mandatory here since we set it in the easyblock
        extra_vars['files_to_copy'][2] = CUSTOM
        return extra_vars

    def __init__(self, *args, **kwargs):
        """Initialize calss variables custom to DualSPHysics."""
        super(EB_DualSPHysics, self).__init__(*args, **kwargs)

        self.dsph_target = None
        self.shortver = '.'.join(self.version.split('.')[0:2])

    def prepare_step(self, *args, **kwargs):
        """Determine name of binary that will be installed."""
        super(EB_DualSPHysics, self).prepare_step(*args, **kwargs)

        if get_software_root('CUDA'):
            self.dsph_target = 'GPU'
        else:
            self.dsph_target = 'CPU'

    def configure_step(self):
        """Custom configure procedure for DualSPHysics."""
        srcdir = os.path.join(self.cfg['start_dir'], 'src/source')
        super(EB_DualSPHysics, self).configure_step(srcdir=srcdir)

    def install_step(self):
        """Custom install procedure for DualSPHysics."""
        # *_linux64 binaries are missing execute permissions
        bindir = os.path.join(self.cfg['start_dir'], 'bin', 'linux')
        for b in glob.glob(os.path.join(bindir, '*_linux64')):
            perms = stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH
            adjust_permissions(b, perms, add=True)

        # no `make install` target
        self.cfg['files_to_copy'] = [
            (['bin/linux/*'], 'bin'),
            (['src/lib/linux_gcc/*'], 'lib'),
        ]
        super(EB_DualSPHysics, self).install_step()

    def post_install_step(self):
        """Custom post-installation step: ensure rpath is patched into binaries/libraries if configured."""
        super(EB_DualSPHysics, self).post_install_step()

        if build_option('rpath'):
            # only the compiled binary (e.g. DualSPHysics5.0CPU_linux64) is rpath'd, the precompiled libraries
            # and binaries are not
            # simple solution: copy the RPATH from the compiled binary to the others, then strip excess paths
            rpathed_bin = os.path.join(
                self.installdir, 'bin', 'DualSPHysics%s%s_linux64' % (self.shortver, self.dsph_target)
            )

            out, _ = run_cmd("patchelf --print-rpath %s" % rpathed_bin, simple=False, trace=False)
            comp_rpath = out.strip()

            files_to_patch = []
            for x in [('bin', '*_linux64'), ('bin', '*.so'), ('lib', '*.so')]:
                files_to_patch.extend(glob.glob(os.path.join(self.installdir, *x)))

            try:
                for x in files_to_patch:
                    out, _ = run_cmd("patchelf --print-rpath %s" % x, trace=False)
                    self.log.debug("Original RPATH for %s: %s" % (out, x))

                    run_cmd("patchelf --set-rpath '%s' --force-rpath %s" % (comp_rpath, x), trace=False)
                    run_cmd("patchelf --shrink-rpath --force-rpath %s" % x, trace=False)

                    out, _ = run_cmd("patchelf --print-rpath %s" % x, trace=False)
                    self.log.debug("RPATH for %s (after patching and shrinking): %s" % (out, x))

            except OSError as err:
                raise EasyBuildError("Failed to patch RPATH section in binaries/libraries: %s", err)

    def sanity_check_step(self):
        """Custom sanity checks for DualSPHysics."""

        # repeated here in case other steps are skipped (e.g. due to --sanity-check-only)
        if get_software_root('CUDA'):
            self.dsph_target = 'GPU'
        else:
            self.dsph_target = 'CPU'

        bins = ['GenCase', 'PartVTK', 'IsoSurface', 'MeasureTool', 'GenCase_MkWord', 'DualSPHysics4.0_LiquidGas',
                'DualSPHysics4.0_LiquidGasCPU', 'DualSPHysics%s' % self.shortver,
                'DualSPHysics%s%s' % (self.shortver, self.dsph_target), 'DualSPHysics%s_NNewtonian' % self.shortver,
                'DualSPHysics%s_NNewtonianCPU' % self.shortver]

        custom_paths = {
            'files': ['bin/%s_linux64' % x for x in bins],
            'dirs': ['lib'],
        }

        custom_commands = ['%s_linux64 -h' % x for x in bins]

        super(EB_DualSPHysics, self).sanity_check_step(custom_paths=custom_paths, custom_commands=custom_commands)
